"""Contain commonly used data structures."""

from functools import lru_cache
from typing import NamedTuple


class DiBool(NamedTuple):
    """A tuple of four bools with directions."""

    top: "bool" = False
    right: "bool" = False
    bottom: "bool" = False
    left: "bool" = False

    @classmethod
    def from_value(cls, value: "bool") -> "DiBool":
        """Construct an instance from a single value."""
        return cls(top=value, right=value, bottom=value, left=value)


class DiInt(NamedTuple):
    """A tuple of four integers with directions."""

    top: "int" = 0
    right: "int" = 0
    bottom: "int" = 0
    left: "int" = 0

    @classmethod
    def from_value(cls, value: "int") -> "DiInt":
        """Construct an instance from a single value."""
        return cls(top=value, right=value, bottom=value, left=value)


class DiStr(NamedTuple):
    """A tuple of four strings with directions."""

    top: "str" = ""
    right: "str" = ""
    bottom: "str" = ""
    left: "str" = ""

    @classmethod
    def from_value(cls, value: "str") -> "DiStr":
        """Construct an instance from a single value."""
        return cls(top=value, right=value, bottom=value, left=value)


class WeightedInt(NamedTuple):
    """Ainterger with an associated weight."""

    weight: "int"
    value: "int"


class WeightedDiInt(NamedTuple):
    """A tuple of four weighted integers."""

    top: "WeightedInt"
    right: "WeightedInt"
    bottom: "WeightedInt"
    left: "WeightedInt"

    # We cannot use :py:func:`functools.cached_property` here as it does not work with
    # :py:Class:`NamedTuple`s.
    @property  # type: ignore
    @lru_cache(maxsize=1)  # noqa: B019
    def unweighted(self) -> "DiInt":
        """Get the padding without weights."""
        return DiInt(*(x.value for x in self))



"""Modify the KeyProcessor to remove any timeout after an escape key press."""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING

from prompt_toolkit.application.current import get_app
from prompt_toolkit.key_binding.key_processor import KeyProcessor as PtKeyProcessor
from prompt_toolkit.key_binding.key_processor import _Flush
from prompt_toolkit.keys import Keys

if TYPE_CHECKING:
    from typing import Any

    from prompt_toolkit.key_binding.key_processor import KeyPress

log = logging.getLogger(__name__)


class KeyProcessor(PtKeyProcessor):
    """A subclass of prompt_toolkit's keyprocessor.

    This adds an exception to the auto-flush timeout so that the input is flushed
    immediately if the key pressed is the escape key.

    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Create a new KeyProcessor instance."""
        super().__init__(*args, **kwargs)
        self._last_key_press: KeyPress | None = None

    def _start_timeout(self) -> None:
        """Start auto flush timeout. Similar to Vim's `timeoutlen` option.

        Do not apply a timeout if the key is :kbd:`Escape`.
        """
        app = get_app()
        timeout = app.timeoutlen

        if timeout is None:
            return

        async def wait() -> None:
            """Wait for timeout."""
            # This sleep can be cancelled. In that case we don't flush.
            # Don't sleep for the escape key
            if self._last_key_press is None or self._last_key_press.key != Keys.Escape:
                await asyncio.sleep(timeout)

            if len(self.key_buffer) > 0:
                # (No keys pressed in the meantime.)
                flush_keys()

        def flush_keys() -> None:
            """Fluh keys."""
            self.feed(_Flush)
            self.process_keys()

        # Automatically flush keys.
        if self._flush_wait_task:
            self._flush_wait_task.cancel()
        self._flush_wait_task = app.create_background_task(wait())

    def process_keys(self) -> None:
        """Process all the keys in the input queue."""
        app = get_app()

        def not_empty() -> bool:
            # When the application result is set, stop processing keys.  (E.g.
            # if ENTER was received, followed by a few additional key strokes,
            # leave the other keys in the queue.)
            if app.is_done:
                # But if there are still CPRResponse keys in the queue, these
                # need to be processed.
                return any(k for k in self.input_queue if k.key == Keys.CPRResponse)
            else:
                return bool(self.input_queue)

        def get_next() -> KeyPress:
            if app.is_done:
                # Only process CPR responses. Everything else is typeahead.
                cpr = [k for k in self.input_queue if k.key == Keys.CPRResponse][0]
                self.input_queue.remove(cpr)
                return cpr
            else:
                return self.input_queue.popleft()

        is_flush = False

        self._last_key_press = None
        while not_empty():
            # Process next key.
            key_press = get_next()
            self._last_key_press = key_press

            is_flush = key_press is _Flush
            is_cpr = key_press.key == Keys.CPRResponse

            if not is_flush and not is_cpr:
                self.before_key_press.fire()

            try:
                self._process_coroutine.send(key_press)
            except Exception:
                # If for some reason something goes wrong in the parser, (maybe
                # an exception was raised) restart the processor for next time.
                self.reset()
                self.empty_queue()
                raise

            if not is_flush and not is_cpr:
                self.after_key_press.fire()

        # Skip timeout if the last key was flush.
        if not is_flush:
            self._start_timeout()



"""Allow drawing tables as :class:`FormattedText`."""

from __future__ import annotations

from collections import defaultdict
from functools import lru_cache, partial
from itertools import tee, zip_longest
from typing import TYPE_CHECKING, cast

from prompt_toolkit.application.current import get_app_session
from prompt_toolkit.formatted_text.base import to_formatted_text
from prompt_toolkit.formatted_text.utils import split_lines, to_plain_text
from prompt_toolkit.layout.dimension import Dimension, to_dimension

from euporie.core.border import (
    DiLineStyle,
    GridChar,
    InvisibleLine,
    LineStyle,
    NoLine,
    ThinLine,
    get_grid_char,
)
from euporie.core.data_structures import (
    DiBool,
    DiInt,
    DiStr,
)
from euporie.core.ft.utils import (
    FormattedTextAlign,
    align,
    fragment_list_width,
    join_lines,
    max_line_width,
    wrap,
)

if TYPE_CHECKING:
    from typing import (
        Any,
        Iterable,
        Iterator,
        Sequence,
        TypeVar,
    )

    from prompt_toolkit.formatted_text.base import (
        AnyFormattedText,
        StyleAndTextTuples,
    )
    from prompt_toolkit.layout.dimension import AnyDimension

    PairT = TypeVar("PairT")


def pairwise(iterable: Iterable[PairT]) -> Iterator[tuple[PairT, PairT]]:
    """Return successiver overlapping pairs from an iterable."""
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


class Cell:
    """A table cell."""

    def __init__(
        self,
        text: AnyFormattedText = "",
        row: Row | None = None,
        col: Col | None = None,
        colspan: int = 1,
        rowspan: int = 1,
        width: int | None = None,
        align: FormattedTextAlign | None = None,
        style: str = "",
        padding: DiInt | int = 0,
        border_line: DiLineStyle | LineStyle = ThinLine,
        border_style: DiStr | str = "",
        border_visibility: DiBool | bool | None = True,
    ) -> None:
        """Create a new table cell.

        Args:
            text: Text or formatted text to display in the cell
            row: The row to which this cell belongs
            col: The column to which this cell belongs
            colspan: The number of columns this cell spans
            rowspan: The number of row this cell spans
            width: The desired width of the cell
            align: How the text in the cell should be aligned
            style: The style to apply to the cell's contents
            padding: The padding around the contents of the cell
            border_line: The line to use for the borders
            border_style: The style to apply to the cell's borders
            border_visibility: The visibility of each border edge

        """
        self.expands = self
        self.text = text
        self.row = row or DummyRow()
        self.col = col or DummyCol()
        self.colspan = colspan
        self.rowspan = rowspan
        self.width = width
        self.align = align
        self.style = style

        self.padding = (
            DiInt(padding, padding, padding, padding)
            if isinstance(padding, int)
            else padding
        )
        self.border_line = (
            DiLineStyle(border_line, border_line, border_line, border_line)
            if isinstance(border_line, LineStyle)
            else border_line
        )
        self.border_style = (
            DiStr(border_style, border_style, border_style, border_style)
            if isinstance(border_style, str)
            else border_style
        )
        self.border_visibility = (
            DiBool(
                border_visibility,
                border_visibility,
                border_visibility,
                border_visibility,
            )
            if isinstance(border_visibility, bool)
            else border_visibility
        )

        self._col_index: int | None = None
        self._row_index: int | None = None

    @property
    def padding(self) -> DiInt:
        """The cell's padding."""
        return self._padding

    @padding.setter
    def padding(self, padding: DiInt | int) -> None:
        """Set the cell's padding."""
        self._padding = (
            DiInt(padding, padding, padding, padding)
            if isinstance(padding, int)
            else padding
        )

    @property
    def border_line(self) -> DiLineStyle:
        """The cell's border line."""
        return self._border_line

    @border_line.setter
    def border_line(self, border_line: DiLineStyle | LineStyle) -> None:
        """Set the cell's border line."""
        self._border_line = (
            DiLineStyle(border_line, border_line, border_line, border_line)
            if isinstance(border_line, LineStyle)
            else border_line
        )

    @property
    def border_style(self) -> DiStr:
        """The cell's border style."""
        return self._border_style

    @border_style.setter
    def border_style(self, border_style: DiStr | str) -> None:
        """Set the cell's border style."""
        self._border_style = (
            DiStr(border_style, border_style, border_style, border_style)
            if isinstance(border_style, str)
            else border_style
        )

    def __repr__(self) -> str:
        """Return a text representation of the cell."""
        cell_text = to_plain_text(self.text)
        if len(cell_text) > 5:
            cell_text = cell_text[:4] + "…"
        return f"{self.__class__.__name__}({cell_text!r})"


class SpacerCell(Cell):
    """A dummy cell to virtually occupy space when ``colspan`` or ``rowspan`` are used."""

    def __init__(
        self,
        expands: Cell,
        span_row_index: int,
        span_col_index: int,
        text: AnyFormattedText = "",
        row: Row | None = None,
        col: Col | None = None,
        colspan: int = 1,
        rowspan: int = 1,
        width: int | None = None,
        align: FormattedTextAlign | None = None,
        style: str = "",
        padding: DiInt | int | None = None,
        border_line: DiLineStyle | LineStyle = NoLine,
        border_style: DiStr | str = "",
        border_visibility: DiBool | bool | None = None,
    ) -> None:
        """Create a new table cell.

        Args:
            expands: This should be reference to the spanning cell
            span_col_index: The index of a spacer cell inside a colspan
            span_row_index: The index of a spacer cell inside a rowspan
            text: Text or formatted text to display in the cell
            row: The row to which this cell belongs
            col: The column to which this cell belongs
            colspan: The number of columns this cell spans
            rowspan: The number of row this cell spans
            width: The desired width of the cell
            align: How the text in the cell should be aligned
            style: The style to apply to the cell's contents
            padding: The padding around the contents of the cell
            border_line: The line to use for the borders
            border_style: The style to apply to the cell's borders
            border_visibility: The visibility of each border edge

        """
        super().__init__(
            text="",
            row=row,
            col=col,
            colspan=1,
            rowspan=1,
            width=None,
            align=align,
            style=expands.style,
            padding=0,
            border_line=expands.border_line,
            border_style=expands.border_style,
            border_visibility=expands.border_visibility,
        )
        self.expands = expands
        self.span_row_index = span_row_index
        self.span_col_index = span_col_index


class _Dummy(Cell):
    """A dummy cell with not content, padding or borders."""

    def __init__(
        self,
        text: AnyFormattedText = "",
        row: Row | None = None,
        col: Col | None = None,
        colspan: int = 1,
        rowspan: int = 1,
        width: int | None = None,
        align: FormattedTextAlign | None = None,
        style: str = "",
        padding: DiInt | int = 0,
        border_line: DiLineStyle | LineStyle = NoLine,
        border_style: DiStr | str = "",
        border_visibility: DiBool | bool | None = None,
    ) -> None:
        """Create an dummy cells to fill empty space in a table."""
        table = (row or col or DummyRow()).table
        background_style = table.background_style
        super().__init__(
            row=row,
            col=col,
            style=background_style,
            padding=0,
            border_line=NoLine,
            border_style=background_style,
            border_visibility=False,
        )

    def __repr__(self) -> str:
        """Represent the cell instance as a string."""
        return f"{self.__class__.__name__}()"


class RowCol:
    """Base class for table rows and columns."""

    _type: str
    span_type: str

    def __init__(
        self,
        table: Table | None = None,
        cells: Sequence[Cell] | None = None,
        align: FormattedTextAlign | None = None,
        style: str = "",
        padding: DiInt | int = 0,
        border_line: DiLineStyle | LineStyle = NoLine,
        border_style: DiStr | str = "",
        border_visibility: DiBool | bool | None = None,
    ) -> None:
        """Create a new row/column.

        Args:
            table: The :py:class:`table` that this row/column belongs to
            cells: A list of cells in this row/column
            align: The default alignment for cells in this row/column
            style: The default style for cells in this row/column
            padding: The default padding for cells in this row/column
            border_line: The line to use for the borders
            border_style: The style to apply to the table's borders
            border_visibility: The visibility of each border edge

        """
        self.table = table or DummyTable()
        if isinstance(self, Row):
            row = self
            col = None
        elif isinstance(self, Col):
            row = None
            col = self
        else:
            raise ValueError("%s is not a `Row` or `Col`", self)

        self._cells = defaultdict(
            lambda: _Dummy(border_style=self.table.style, row=row, col=col),
            dict(enumerate(cells or [])),
        )
        if cells:
            for cell in cells:
                setattr(cell, self._type, self)

        self.align = align
        self.style = style

        self._padding = (
            DiInt(padding, padding, padding, padding)
            if isinstance(padding, int)
            else padding
        )
        self._border_line = (
            DiLineStyle(border_line, border_line, border_line, border_line)
            if isinstance(border_line, LineStyle)
            else border_line
        )
        self._border_style = (
            DiStr(border_style, border_style, border_style, border_style)
            if isinstance(border_style, str)
            else border_style
        )
        self.border_visibility = (
            DiBool(
                border_visibility,
                border_visibility,
                border_visibility,
                border_visibility,
            )
            if isinstance(border_visibility, bool)
            else border_visibility
        )

    @property
    def padding(self) -> DiInt:
        """The cell's padding."""
        return self._padding

    @padding.setter
    def padding(self, padding: DiInt | int) -> None:
        """Set the cell's padding."""
        self._padding = (
            DiInt(padding, padding, padding, padding)
            if isinstance(padding, int)
            else padding
        )

    @property
    def border_line(self) -> DiLineStyle:
        """The cell's border line."""
        return self._border_line

    @border_line.setter
    def border_line(self, border_line: DiLineStyle | LineStyle) -> None:
        """Set the cell's border line."""
        self._border_line = (
            DiLineStyle(border_line, border_line, border_line, border_line)
            if isinstance(border_line, LineStyle)
            else border_line
        )

    @property
    def border_style(self) -> DiStr:
        """The cell's border style."""
        return self._border_style

    @border_style.setter
    def border_style(self, border_style: DiStr | str) -> None:
        """Set the cell's border style."""
        self._border_style = (
            DiStr(border_style, border_style, border_style, border_style)
            if isinstance(border_style, str)
            else border_style
        )

    @property
    def cells(self) -> list[Cell]:
        """List the cells in the row/column."""
        if self._type == "row":
            n = len(self.table._cols)
        else:
            n = len(self.table._rows)
        cells = []
        for i in range(n):
            cell = self._cells[i]
            if isinstance(cell, _Dummy):
                if isinstance(self, Row):
                    cell.row = self
                    cell.col = self.table._cols[i]
                elif isinstance(self, Col):
                    cell.col = self
                    cell.row = self.table._rows[i]
            cells.append(cell)

        return cells

    def new_cell(self, *args: Any, **kwargs: Any) -> Cell:
        """Create a new cell in this row/column."""
        cell = Cell(*args, **kwargs)
        self.add_cell(cell)
        return cell

    def add_cell(self, cell: Cell, index: int | None = None) -> None:
        """Add a cell to the row/ column."""
        if index is None:
            # Fit cells into available space
            # TODO - also check fit by colspan
            index = 0
            cells = self._cells
            rowspan = cell.rowspan
            colspan = cell.colspan
            while any(
                index + y in cells
                for y in range(rowspan if self._type == "col" else colspan)
            ):
                index += 1

        self._cells[index] = cell

        if self._type == "row":
            assert isinstance(self, Row)
            cell.row = cast("Row", self)
            cell.col = self.table._cols[index]

            row_index = next(i for i, row in self.table._rows.items() if row is self)
            col_index = index

            cell.col._cells[row_index] = cell

            cell._row_index = row_index
            cell._col_index = index

        elif self._type == "col":
            assert isinstance(self, Col)
            cell.row = self.table._rows[index]
            cell.col = cast("Col", self)

            row_index = index
            col_index = next(i for i, col in self.table._cols.items() if col is self)

            cell.row._cells[col_index] = cell

            cell._row_index = index
            cell._col_index = col_index

        for i in range(cell.rowspan):
            for j in range(cell.colspan):
                if i > 0 or j > 0:
                    spacer = SpacerCell(
                        expands=cell,
                        span_row_index=i,
                        span_col_index=j,
                        row=self.table._rows[row_index + i],
                        col=self.table._cols[col_index + j],
                    )
                    spacer._row_index = row_index + i
                    spacer._col_index = col_index + j
                    self.table._rows[row_index + i]._cells[col_index + j] = spacer
                    self.table._cols[col_index + j]._cells[row_index + i] = spacer

    def __repr__(self) -> str:
        """Return a textual representation of the row or column."""
        return f"{self.__class__.__name__}({', '.join(map(str, self._cells.values()))})"


class Row(RowCol):
    """A row in a table."""

    _type = "row"
    span_type = "colspan"


class Col(RowCol):
    """A column in a table."""

    _type = "col"
    span_type = "rowspan"


@lru_cache
def compute_style(cell: Cell, render_count: int = 0) -> str:
    """Compute a cell's style string."""
    return f"{cell.row.table.style} {cell.col.style} {cell.row.style} {cell.style}"


###############


@lru_cache
def compute_text(cell: Cell, render_count: int = 0) -> StyleAndTextTuples:
    """Compute a cell's input, converted to :class:`FormattedText`."""
    return to_formatted_text(cell.text, style=compute_style(cell))


@lru_cache
def compute_padding(cell: Cell, render_count: int = 0) -> DiInt:
    """Compute a cell's padding."""
    output = {}
    table_padding = cell.row.table.padding
    row_padding = cell.row.padding
    col_padding = cell.col.padding
    cell_padding = cell.padding
    for direction in ("top", "right", "bottom", "left"):
        output[direction] = max(
            getattr(table_padding, direction, 0),
            getattr(row_padding, direction, 0),
            getattr(col_padding, direction, 0),
            getattr(cell_padding, direction, 0),
            0,
        )
    return DiInt(**output)


@lru_cache
def compute_border_visibility(cell: Cell, render_count: int = 0) -> DiBool:
    """Compute a cell's border visibility."""
    output = {}

    row = cell.row
    col = cell.col
    table = row.table
    n_rows = len(table._rows)
    n_cols = len(table._cols)
    cell_border_visibility = cell.border_visibility
    row_border_visibility = row.border_visibility
    col_border_visibility = col.border_visibility
    table_border_visibility = table.border_visibility

    output["top"] = (
        (cell_border_visibility and cell_border_visibility.top)
        or (row_border_visibility and row_border_visibility.top)
        or (
            cell._row_index == 0 and col_border_visibility and col_border_visibility.top
        )
        or (
            cell._row_index == 0
            and table_border_visibility
            and table_border_visibility.top
        )
    )

    output["bottom"] = (
        (cell_border_visibility and cell_border_visibility.bottom)
        or (row_border_visibility and row_border_visibility.bottom)
        or (
            cell._row_index == n_rows - 1
            and col_border_visibility
            and col_border_visibility.bottom
        )
        or (
            table_border_visibility
            and cell._row_index == n_rows - 1
            and table_border_visibility.bottom
        )
    )

    output["left"] = (
        (cell_border_visibility and cell_border_visibility.left)
        or (
            cell._col_index == 0
            and row_border_visibility
            and row_border_visibility.left
        )
        or (col_border_visibility and col_border_visibility.left)
        or (
            cell._col_index == 0
            and table_border_visibility
            and table_border_visibility.left
        )
    )

    output["right"] = (
        (cell_border_visibility and cell_border_visibility.right)
        or (
            cell._col_index == n_cols - 1
            and row_border_visibility
            and row_border_visibility.left
        )
        or (col_border_visibility and col_border_visibility.left)
        or (
            table_border_visibility
            and cell._col_index == n_cols - 1
            and table_border_visibility.right
        )
    )

    return DiBool(**output)


###############


@lru_cache
def calculate_cell_width(cell: Cell, render_count: int = 0) -> int:
    """Compute the final width of a cell, including padding."""
    if cell.colspan > 1:
        return 0
    padding = compute_padding(cell, render_count)
    return (
        (cell.width or max_line_width(compute_text(cell, render_count)))
        + padding.left
        + padding.right
    )


@lru_cache
def compute_border_line(cell: Cell, render_count: int = 0) -> DiLineStyle:
    """Compute a cell's border line."""
    if isinstance(cell, _Dummy):
        return DiLineStyle(NoLine, NoLine, NoLine, NoLine)
    output = {}

    row = cell.row
    col = cell.col
    table = cell.row.table
    row_cells = row.cells
    col_cells = col.cells
    cell_border_line = cell.border_line
    row_border_line = row.border_line
    col_border_line = col.border_line
    table_border_line = table.border_line

    if (value := row_border_line.top) is not NoLine:
        output["top"] = value
    elif cell == col_cells[0] and (value := table_border_line.top) is not NoLine:
        output["top"] = value

    if (value := row_border_line.bottom) is not NoLine:
        output["bottom"] = value
    elif cell == col_cells[-1] and (value := table_border_line.bottom) is not NoLine:
        output["bottom"] = value

    if (value := col_border_line.left) is not NoLine:
        output["left"] = value
    elif cell == row_cells[0] and (value := table_border_line.left) is not NoLine:
        output["left"] = value

    if (value := col_border_line.right) is not NoLine:
        output["right"] = value
    elif cell == row_cells[-1] and (value := table_border_line.right) is not NoLine:
        output["right"] = value

    for direction in ("top", "right", "bottom", "left"):
        if (value := getattr(cell_border_line, direction, NoLine)) is not NoLine:
            output[direction] = value

    if isinstance(cell, SpacerCell):
        expands = cell.expands
        # Set left border to invisible in colspan
        if 0 < cell.span_col_index < expands.colspan:
            output["left"] = InvisibleLine
        # Set top border to invisible in rowspan
        if 0 < cell.span_row_index < expands.rowspan:
            output["top"] = InvisibleLine

    return DiLineStyle(**output)


@lru_cache
def compute_border_width(cell: Cell, render_count: int = 0) -> DiInt:
    """Compute the width of a cell's borders."""
    if isinstance(cell, _Dummy):
        output = {"top": 0, "right": 0, "bottom": 0, "left": 0}
    else:
        output = {"top": 1, "right": 1, "bottom": 1, "left": 1}

    # XXX If don't understand why row needs to follow cell expansion but column doesn't
    row = cell.row
    col = cell.col
    table = row.table

    if row not in table._rows.values() or col not in table._cols.values():
        # The table has not yet been fully constructed
        return DiInt(**output)

    row_index = table.rows.index(row)
    col_index = table.cols.index(col)

    row_cells_bv = [compute_border_visibility(cell, render_count) for cell in row.cells]
    col_cells_bv = [compute_border_visibility(cell, render_count) for cell in col.cells]

    for cell, cell_left in zip(row.cells, table.rows[row_index - 1].cells):
        if (
            compute_border_visibility(cell).left
            or compute_border_visibility(cell_left).right
        ):
            break

    row_top_cells_bv = (
        compute_border_visibility(cell, render_count)
        for cell in (
            table.rows[row_index - 1].cells if row_index - 1 in table._rows else []
        )
    )
    col_left_cells_bv = (
        compute_border_visibility(cell, render_count)
        for cell in (
            table.cols[col_index - 1].cells if col_index - 1 in table._cols else []
        )
    )
    col_right_cells_bv = (
        compute_border_visibility(cell, render_count)
        for cell in (
            table.cols[col_index + 1].cells if col_index + 1 in table._cols else []
        )
    )
    row_bottom_cells_bv = (
        compute_border_visibility(cell, render_count)
        for cell in (
            table.rows[row_index + 1].cells if row_index + 1 in table._rows else []
        )
    )

    output["top"] = int(
        any(
            bv.top or (bv_other and bv_other.bottom)
            for bv, bv_other in zip_longest(row_cells_bv, row_top_cells_bv)
        )
    )
    output["right"] = int(
        any(
            bv.right or (bv_other and bv_other.left)
            for bv, bv_other in zip_longest(col_cells_bv, col_right_cells_bv)
        )
    )
    output["bottom"] = int(
        any(
            bv.bottom or (bv_other and bv_other.top)
            for bv, bv_other in zip_longest(row_cells_bv, row_bottom_cells_bv)
        )
    )
    output["left"] = int(
        any(
            bv.left or (bv_other and bv_other.right)
            for bv, bv_other in zip_longest(col_cells_bv, col_left_cells_bv)
        )
    )

    return DiInt(**output)


@lru_cache
def calculate_col_widths(
    cols: tuple[Col],
    width: Dimension,
    expand_to_width: bool,
    min_col_width: int = 2,
    render_count: int = 0,
) -> list[int]:
    """Calculate column widths given the available space.

    Reduce the widest column until we fit in available width, or expand cells to
    to fill the available width.

    Args:
        cols: A list of columns in the table
        width: The desired width of the table
        expand_to_width: Whether the column should expand to fill the available width
        min_col_width: The minimum width allowed for a column
        render_count: The number of times the app has been rendered

    Returns:
        List of new column widths

    """
    # TODO - this function is too slow

    col_widths = [
        max(
            min_col_width,
            *(calculate_cell_width(cell, render_count) for cell in col.cells),
        )
        for col in cols
    ]

    def total_width(col_widths: list[int]) -> int:
        """Calculate the total width of the columns including borders."""
        width = sum(col_widths)
        if cols:
            width += compute_border_width(cols[0].cells[0], render_count).left
        for _i, col in enumerate(cols):
            width += compute_border_width(col.cells[0], render_count).right
        return width

    def expand(target: int) -> None:
        """Expand the columns."""
        max_width = max(target, len(col_widths) * min_col_width)
        while total_width(col_widths) < max_width:
            # Expand only columns which do not have a width set if possible
            # TODO - expand proportionately to given widths
            col_index_widths = [
                (i, col_widths[i])
                for i, col in enumerate(cols)
                if all(cell.width is None for cell in col.cells)
            ]
            if not col_index_widths:
                col_index_widths = list(enumerate(col_widths))
            if col_index_widths:
                idxmin = min(col_index_widths, key=lambda x: x[1])[0]
                col_widths[idxmin] += 1
            else:
                break

    def contract(target: int) -> None:
        """Contract the columns."""
        max_width = max(target, len(col_widths) * min_col_width)
        while total_width(col_widths) > max_width:
            idxmax = max(enumerate(col_widths), key=lambda x: x[1])[0]
            col_widths[idxmax] -= 1

    # Determine whether to expand or contract the table
    current_width = total_width(col_widths)
    if width.preferred_specified:
        if current_width < width.preferred:
            expand(width.preferred)
        elif current_width > width.preferred:
            contract(width.preferred)
    if width.max_specified:
        if current_width > width.max:
            contract(width.max)
        if current_width < width.max and expand_to_width:
            expand(width.max)
    if width.min_specified and current_width < width.min:
        expand(width.min)

    return col_widths


@lru_cache
def get_node(
    nw_bl: DiLineStyle,
    ne_bl: DiLineStyle,
    se_bl: DiLineStyle,
    sw_bl: DiLineStyle,
) -> str:
    """Calculate which character to use at the intersection of four cells."""
    return get_grid_char(
        GridChar(
            max(nw_bl.right, ne_bl.left),
            max(ne_bl.bottom, se_bl.top),
            max(se_bl.left, sw_bl.right),
            max(sw_bl.top, nw_bl.bottom),
        )
    )


@lru_cache
def get_horizontal_edge(n_bl: DiLineStyle, s_bl: DiLineStyle) -> str:
    """Calculate which character to use to divide horizontally adjacent cells."""
    line_style = max(n_bl.bottom, s_bl.top)
    return get_grid_char(GridChar(NoLine, line_style, NoLine, line_style))


@lru_cache
def get_vertical_edge(w_bl: DiLineStyle, e_bl: DiLineStyle) -> str:
    """Calculate which character to use to divide vertically adjacent cells."""
    line_style = max(w_bl.right, e_bl.left)
    return get_grid_char(GridChar(line_style, NoLine, line_style, NoLine))


@lru_cache
def compute_align(cell: Cell, render_count: int = 0) -> FormattedTextAlign:
    """Compute the alignment of a cell."""
    if (align := cell.align) is not None:
        return align
    else:
        return cell.row.align or cell.col.align or cell.row.table.align


@lru_cache
def compute_lines(
    cell: Cell, width: int, render_count: int = 0
) -> list[StyleAndTextTuples]:
    """Wrap the cell's text to a given width.

    Args:
        cell: The cell whose lines to compute
        width: The width at which to wrap the cell's text.
        render_count: The number of times the application has been rendered

    Returns:
        A list of lines of formatted text
    """
    padding = compute_padding(cell, render_count)
    return list(
        split_lines(
            align(
                wrap(
                    [
                        *([("", "\n" * padding.top)] if padding.top else []),
                        *compute_text(cell, render_count),
                        *([("", "\n" * padding.bottom)] if padding.bottom else []),
                    ],
                    width=width,
                    placeholder="",
                ),
                compute_align(cell, render_count),
                width=width,
                style=compute_style(cell, render_count),
                placeholder="",
            )
        )
    )


@lru_cache
def compute_border_style(cell: Cell, render_count: int = 0) -> DiStr:
    """Compute the cell's final style for each of a cell's borders."""
    output = {}

    row = cell.row
    col = cell.col
    table = row.table
    max_row = len(table._rows) - 1
    max_col = len(table._cols) - 1
    cell_border_style = cell.border_style
    row_border_style = row.border_style
    col_border_style = col.border_style
    table_border_style = table.border_style

    output["top"] = "".join(
        (
            table_border_style.top if cell._row_index == 0 else "",
            col_border_style.top if cell._row_index == 0 else "",
            row_border_style.top,
            cell_border_style.top,
        )
    )
    output["right"] = "".join(
        (
            table_border_style.right if cell._col_index == max_col else "",
            col_border_style.right,
            row_border_style.right if cell._col_index == max_col else "",
            cell_border_style.right,
        )
    )
    output["bottom"] = "".join(
        (
            table_border_style.bottom if cell._row_index == max_row else "",
            col_border_style.bottom if cell._row_index == max_row else "",
            row_border_style.bottom,
            cell_border_style.bottom,
        )
    )
    output["left"] = "".join(
        (
            table_border_style.left if cell._col_index == 0 else "",
            col_border_style.left,
            row_border_style.left if cell._col_index == 0 else "",
            cell_border_style.left,
        )
    )

    return DiStr(**output)


class Table:
    """A table."""

    def __init__(
        self,
        rows: Sequence[Row] | None = None,
        cols: Sequence[Col] | None = None,
        width: AnyDimension | None = None,
        expand: bool = False,
        align: FormattedTextAlign = FormattedTextAlign.LEFT,
        style: str = "",
        padding: DiInt | int | None = None,
        border_line: DiLineStyle | LineStyle = NoLine,
        border_style: DiStr | str = "",
        border_visibility: DiBool | bool = False,
        background_style: str = "",
    ) -> None:
        """Create a new table instance.

        Args:
            rows: A list of :class:`Row`s to add to the table
            cols: A list of :class:`Col`s to add to the table. Cells specified in rows
                take priority, so cells in column positions already defined by rows
                will be ignored.
            width: The width of the table
            expand: Whether the table should expand to fill the available space
            align: The default alignment for cells in the table
            style: A style to apply to the table's cells' contents
            padding: The default padding for cells in the table
            border_line: The line to use for the borders
            border_style: The style to apply to the table's borders
            border_visibility: If :const:`True`, if borders in the table are
                :class:`Invisible` for their entire length, no extra line will be drawn
            background_style: The style to apply to missing cells

        """
        self.render_count = 0

        self._rows = defaultdict(partial(Row, self), enumerate(rows or []))
        if rows:
            for row in rows:
                row.table = self
        self._cols = defaultdict(partial(Col, self), enumerate(cols or []))
        if cols:
            for col in cols:
                col.table = self
        if rows:
            self.sync_rows_to_cols()
        elif cols:
            self.sync_cols_to_rows()

        if width is None:
            width = Dimension(max=get_app_session().output.get_size()[1])
        self._width = to_dimension(width)
        self.expand = expand

        self.align = align
        self.style = style

        if padding is None:
            padding = DiInt(0, 1, 0, 1)
        self.padding = (
            DiInt(padding, padding, padding, padding)
            if isinstance(padding, int)
            else padding
        )
        self.border_line = (
            DiLineStyle(border_line, border_line, border_line, border_line)
            if isinstance(border_line, LineStyle)
            else border_line
        )
        self.border_style = (
            DiStr(border_style, border_style, border_style, border_style)
            if isinstance(border_style, str)
            else border_style
        )
        self.border_visibility: DiBool = (
            DiBool(
                border_visibility,
                border_visibility,
                border_visibility,
                border_visibility,
            )
            if isinstance(border_visibility, bool)
            else border_visibility
        )
        self.background_style: str = background_style

    @property
    def padding(self) -> DiInt:
        """The cell's padding."""
        return self._padding

    @padding.setter
    def padding(self, padding: DiInt | int) -> None:
        """Set the cell's padding."""
        self._padding = (
            DiInt(padding, padding, padding, padding)
            if isinstance(padding, int)
            else padding
        )

    @property
    def border_line(self) -> DiLineStyle:
        """The cell's border line."""
        return self._border_line

    @border_line.setter
    def border_line(self, border_line: DiLineStyle | LineStyle) -> None:
        """Set the cell's border line."""
        self._border_line = (
            DiLineStyle(border_line, border_line, border_line, border_line)
            if isinstance(border_line, LineStyle)
            else border_line
        )

    @property
    def border_style(self) -> DiStr:
        """The cell's border style."""
        return self._border_style

    @border_style.setter
    def border_style(self, border_style: DiStr | str) -> None:
        """Set the cell's border style."""
        self._border_style = (
            DiStr(border_style, border_style, border_style, border_style)
            if isinstance(border_style, str)
            else border_style
        )

    @property
    def width(self) -> Dimension:
        """The table's width."""
        return self._width

    @width.setter
    def width(self, value: AnyDimension) -> None:
        """Set the table's width."""
        self._width = to_dimension(value)

    @property
    def rows(self) -> list[Row]:
        """A list of rows in the table."""
        return [self._rows[i] for i in range(len(self._rows))]

    @property
    def cols(self) -> list[Col]:
        """A list of columns in the table."""
        return [self._cols[i] for i in range(len(self._cols))]

    def sync_rows_to_cols(self) -> None:
        """Enure cells in rows are present in the relevant columns."""
        cols = self._cols
        for i, row in self._rows.items():
            for j, cell in row._cells.items():
                cell.col = cols[j]
                cols[j]._cells[i] = cell

    def sync_cols_to_rows(self) -> None:
        """Enure cells in columns are present in the relevant rows."""
        rows = self._rows
        for i, col in self._cols.items():
            for j, cell in col._cells.items():
                cell.row = rows[j]
                rows[j]._cells[i] = cell

    def new_row(self, *args: Any, **kwargs: Any) -> Row:
        """Create a new row in the table."""
        row = Row(*args, **kwargs)
        self.add_row(row)
        return row

    def add_row(self, row: Row) -> None:
        """Add a row to the table."""
        row.table = self

        unfilled = [
            i
            for i, row in self._rows.items()
            if all(isinstance(cell, SpacerCell) for cell in row._cells.values())
        ]

        index = min(
            min(unfilled + [len(self._rows)]),
            max([-1] + list(self._rows)) + 1,
        )

        # Merge existing row with new row
        cells = self._rows[index]._cells
        # Update column reference on existing cells
        for cell in cells.values():
            cell.row = row
        # Add cells from new row to the current cell list
        i = 0
        for cell in row._cells.values():
            while i in cells:
                i += 1
            cell._col_index = i
            cells[i] = cell
        # Update the new row's cell list
        row._cells = cells
        # Add the new row to the table
        self._rows[index] = row

    def new_col(self, *args: Any, **kwargs: Any) -> Col:
        """Create a new column in the table."""
        col = Col(*args, **kwargs)
        self.add_col(col)
        return col

    def add_col(self, col: Col) -> None:
        """Add a column to the table."""
        col.table = self

        unfilled = [
            i
            for i, col in self._cols.items()
            if all(isinstance(cell, SpacerCell) for cell in col._cells.values())
        ]

        index = min(
            min(unfilled + [len(self._cols)]),
            max([-1] + list(self._cols)) + 1,
        )

        # Merge existing col with new col
        cells = self._cols[index]._cells
        # Update column reference on existing cells
        for cell in cells.values():
            cell.col = col
        # Add cells from new column to the current cell list
        i = 0
        for cell in col._cells.values():
            while i in cells:
                i += 1
            cell._row_index = i
            cells[i] = cell
        # Update the new columns's cell list
        col._cells = cells
        # Add the new column to the table
        self._cols[index] = col

    def calculate_col_widths(
        self,
        width: AnyDimension | None = None,
        min_col_width: int = 4,
    ) -> list[int]:
        """Calculate the table's column widths."""
        width = self.width if width is None else to_dimension(width)
        return calculate_col_widths(
            tuple(self.cols), width, self.expand, self.render_count
        )

    def calculate_cell_widths(
        self, width: AnyDimension | None = None
    ) -> dict[Cell, int]:
        """Calculate widths for each table cell, taking colspans into account."""
        render_count = self.render_count
        col_widths = self.calculate_col_widths(width)
        widths = {}
        for _y, row in enumerate(self.rows):
            for x, cell in enumerate(row.cells):
                widths[cell] = 0
                if cell.expands == cell:
                    colspan = cell.colspan
                    widths[cell] += sum(col_widths[x : x + colspan]) + sum(
                        compute_border_width(w, render_count).right
                        or compute_border_width(e, render_count).left
                        for w, e in pairwise(row.cells[x : x + colspan])
                    )
        return widths

    def draw_table_row(
        self,
        row_above: RowCol | None,
        row_below: RowCol | None,
        cell_widths: dict[Cell, int],
        col_widths: list[int],
        row_edge_visibility: bool,
        col_edge_visibilities: dict[int, bool],
    ) -> Iterable[StyleAndTextTuples]:
        """Draw a row in the table."""

        def _calc_cell_inner_width(cell: Cell) -> int:
            """Calculate a cell's inner width."""
            if isinstance(cell, SpacerCell) and cell.span_col_index > 0:
                return 0
            cell = cell.expands
            # Sum widths of spanned columns
            width = cell_widths[cell]
            # Remove padding
            padding = compute_padding(cell, render_count)
            width -= (padding.left or 0) + (padding.right or 0)
            return max(width, 0)

        #################

        assert row_above is not None or row_below is not None

        render_count = self.render_count
        dummy = _Dummy()

        row_above_cells: list[Cell]
        row_below_cells: list[Cell]
        if row_above is None:
            assert row_below is not None
            row_above_cells = [dummy] * len(row_below.cells)
        else:
            row_above_cells = row_above.cells
        if row_below is None:
            assert row_above is not None
            row_below_cells = [dummy] * len(row_above.cells)
        else:
            row_below_cells = row_below.cells
        row_above_cells_pad = [dummy, *row_above_cells, dummy]
        row_below_cells_pad = [dummy, *row_below_cells, dummy]

        if row_above is not None:
            # Skip rows where all cells expand a cell above
            # if all(
            #     isinstance(cell, SpacerCell) and cell.span_row_index > 0
            #     for cell in row_above_cells
            # ):
            #     return output_lines

            # Calculate borders
            borders = []
            render_count = self.render_count
            for i, (w, e) in enumerate(pairwise(row_above_cells_pad)):
                # We only need to check on cell to the left and one cell to the right
                if not col_edge_visibilities[i]:
                    borders.append(("", ""))

                else:
                    border_style = " ".join(
                        (
                            compute_border_style(e, render_count).left,
                            compute_border_style(w, render_count).right,
                        )
                    )
                    border_char = get_vertical_edge(
                        compute_border_line(w, render_count),
                        compute_border_line(e, render_count),
                    )
                    borders.append((border_style, border_char))

            # Draw row contents line by line

            # Calculate the lines in each cell.
            # The result of `compute_lines` is memoized, so the same list instance gets
            # returned each time we call this.
            # We will remove rows from the returned instance of each cell's line list
            # to keep track of remaining lines for row-span cells
            row_cell_lines = [
                compute_lines(
                    cell.expands,
                    width=_calc_cell_inner_width(cell),
                    render_count=render_count,
                )
                for col, cell in enumerate(row_above_cells)
            ]

            # We will iterate over a copy of each cell's lines, as we are removing
            # items from the list as we go
            row_lines = zip_longest(*(x[:] for x in row_cell_lines))
            for row_line in row_lines:
                output_line: StyleAndTextTuples = []

                # If we are not on the last row of a rowspan, carry lines over to the
                # next table row
                if all(
                    line is None
                    or (cell.rowspan > 1)
                    or (
                        isinstance(cell, SpacerCell)
                        and cell.expands.rowspan > 1
                        and (cell.span_row_index > 1 or cell.span_col_index)
                        and cell.span_row_index < cell.expands.rowspan - 1
                    )
                    or (
                        isinstance(cell, SpacerCell)
                        and cell.expands.colspan > 1
                        and cell.span_col_index
                    )
                    for line, cell in zip(row_line, row_above_cells)
                ):
                    break

                for i, (cell_line, cell, row_cell_line) in enumerate(
                    zip(row_line, row_above_cells, row_cell_lines)
                ):
                    # Skip horizontal spacer cells
                    if isinstance(cell, SpacerCell) and (cell.span_col_index):
                        continue

                    output_line.append(borders[i])

                    cell_style = compute_style(cell, render_count)
                    padding_style = f"{cell_style} nounderline"
                    padding = compute_padding(cell, render_count)
                    padding_left, padding_right = padding.left, padding.right

                    excess = _calc_cell_inner_width(cell) - fragment_list_width(
                        cell_line or []
                    )

                    output_line.extend(
                        [
                            (padding_style, " " * (padding_left or 0)),
                            *(cell_line or []),
                            (cell_style + " nounderline", " " * excess),
                            (padding_style, " " * (padding_right or 0)),
                        ]
                    )

                    if row_cell_line:
                        row_cell_line.pop(0)

                output_line.append(borders[i + 1])
                yield output_line

        ##################################
        # Draw border row below
        ##################################

        # Don't draw border rows if all cells expand a cell above
        if row_below and all(
            isinstance(cell, SpacerCell) and cell.span_row_index > 0
            for cell in row_below_cells
        ):
            return

        # Draw a border row if at least once cell above have a visible bottom border
        # or a cell below has a visible top border
        if row_edge_visibility:
            output_line = []

            for i, ((nw, ne), (sw, se)) in enumerate(
                zip(
                    pairwise(row_above_cells_pad),
                    pairwise(row_below_cells_pad),
                )
            ):
                cell = se

                # Skip horizontal spacer cells
                if isinstance(cell, SpacerCell):
                    if cell.span_col_index and cell.span_row_index:
                        continue

                nw_bs = compute_border_style(nw, render_count)
                ne_bs = compute_border_style(ne, render_count)
                se_bs = compute_border_style(se, render_count)
                sw_bs = compute_border_style(sw, render_count)

                if sw.expands == se.expands == nw.expands == ne.expands:
                    node_style = sw.expands.style
                else:
                    # TODO - use style where most directions agree
                    node_style = " ".join(
                        (
                            se_bs.left,
                            se_bs.top,
                            ne_bs.bottom,
                            ne_bs.left,
                            sw_bs.right,
                            sw_bs.top,
                            nw_bs.bottom,
                            nw_bs.right,
                        )
                    )
                # Do not draw border nodes if one dimension is not visible
                ne_bl = compute_border_line(ne, render_count)
                se_bl = compute_border_line(se, render_count)

                if not row_edge_visibility or not col_edge_visibilities[i]:
                    node_char = ""
                else:
                    node_char = get_node(
                        compute_border_line(nw, render_count),
                        ne_bl,
                        se_bl,
                        compute_border_line(sw, render_count),
                    )

                if node_char:
                    output_line.append((node_style, node_char))

                if i < len(row_above_cells):
                    if isinstance(cell, SpacerCell) and cell.span_row_index:
                        if remaining_lines := row_cell_lines[i]:
                            line = remaining_lines.pop(0)
                        else:
                            line = []
                        cell_style = compute_style(cell, render_count)
                        padding_style = f"{cell_style} nounderline"
                        padding = compute_padding(cell, render_count)
                        padding_left, padding_right = padding.left, padding.right
                        excess = _calc_cell_inner_width(cell) - fragment_list_width(
                            line or []
                        )
                        output_line.extend(
                            [
                                (padding_style, " " * (padding_left or 0)),
                                *line,
                                (cell_style, " " * excess),
                                (padding_style, " " * (padding_right or 0)),
                            ]
                        )
                        continue

                    style = " ".join((se_bs.top, ne_bs.bottom))
                    edge = get_horizontal_edge(ne_bl, se_bl)
                    text = edge * col_widths[i]
                    output_line += [(style, text)]

            yield output_line

    def render(self, width: AnyDimension | None = None) -> StyleAndTextTuples:
        """Draw the table, optionally at a given character width."""
        self.render_count += 1
        width = self.width if width is None else to_dimension(width)

        # Calculate border visibility
        render_count = self.render_count
        row_edge_visibilities: dict[int, bool] = defaultdict(lambda: False)
        col_edge_visibilities: dict[int, bool] = defaultdict(lambda: False)
        for y, row in enumerate(self.rows):
            for x, cell in enumerate(row.cells):
                bv = compute_border_visibility(cell, render_count)
                if bv.left:
                    col_edge_visibilities[x] = True
                if bv.right:
                    col_edge_visibilities[x + 1] = True
                if bv.top:
                    row_edge_visibilities[y] = True
                if bv.bottom:
                    row_edge_visibilities[y + 1] = True

        col_widths = self.calculate_col_widths(width)
        cell_widths = self.calculate_cell_widths(width)

        lines: list[StyleAndTextTuples] = []

        if self.rows:
            for i, (row_above, row_below) in enumerate(
                pairwise([None, *self.rows, None])
            ):
                lines.extend(
                    self.draw_table_row(
                        row_above,
                        row_below,
                        cell_widths,
                        col_widths,
                        row_edge_visibilities[i],
                        col_edge_visibilities,
                    )
                )

        return join_lines(lines)

    def __pt_formatted_text__(self) -> StyleAndTextTuples:
        """Render the table as formatted text."""
        return self.render()


class DummyRow(Row):
    """A dummy row - created to hold cells without an assigned column."""

    def add_cell(self, cell: Cell, index: int | None = None) -> None:
        """Prevent cells being added to a dummy row."""
        raise NotImplementedError("Cannot add a cell to a DummyRow")


class DummyCol(Col):
    """A dummy column - created to hold cells without an assigned column."""

    def add_cell(self, cell: Cell, index: int | None = None) -> None:
        """Prevent cells being added to a dummy column."""
        raise NotImplementedError("Cannot add a cell to a DummyCol")


class DummyTable(Table):
    """A dummy table - created to hold rows and columns without an assigned table."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Create a new dummy table."""
        kwargs["border_line"] = NoLine
        super().__init__(*args, **kwargs)

    def add_row(self, row: Row) -> None:
        """Prevent rows being added to a dummy table."""
        raise NotImplementedError("Cannot add a row to a DummyTable")

    def add_col(self, col: Col) -> None:
        """Prevent columns being added to a dummy table."""
        raise NotImplementedError("Cannot add a column to a DummyTable")



"""Contain classes related to querying terminal features."""

from __future__ import annotations

import array
import logging
import os
import re
import shlex
import subprocess
from functools import lru_cache
from typing import TYPE_CHECKING

from aenum import extend_enum
from prompt_toolkit.application.current import get_app
from prompt_toolkit.application.run_in_terminal import run_in_terminal
from prompt_toolkit.keys import Keys
from prompt_toolkit.output import ColorDepth
from prompt_toolkit.utils import Event

from euporie.core.commands import add_cmd
from euporie.core.filters import in_tmux
from euporie.core.key_binding.registry import register_bindings
from euporie.core.style import DEFAULT_COLORS

if TYPE_CHECKING:
    from typing import Any

    from prompt_toolkit.input import Input
    from prompt_toolkit.key_binding import KeyPressEvent
    from prompt_toolkit.output import Output

    from euporie.core.config import Config

log = logging.getLogger(__name__)


@lru_cache
def _have_termios_tty_fcntl() -> bool:
    try:
        import fcntl  # noqa F401
        import termios  # noqa F401
        import tty  # noqa F401
    except ModuleNotFoundError:
        return False
    else:
        return True


def tmuxify(cmd: str) -> str:
    """Wrap an escape sequence for tmux passthrough."""
    if in_tmux():
        cmd = cmd.replace("\x1b", "\x1b\x1b")
        cmd = f"\x1bPtmux;{cmd}\033\\"
    return cmd


class TerminalQuery:
    """A class representing a terminal query.

    This allows a control sequence to sent to the terminal, the response interpreted,
    and the received value processed and stored.
    """

    default: Any | None = None
    cache = False
    cmd = ""
    pattern: re.Pattern | None = None

    def __init__(self, output: Output, config: Config) -> None:
        """Create a new instance of the terminal query."""
        self.config = config
        self.output = output
        self.waiting = False
        self._value: Any | None = None
        self.event = Event(self)
        self.queryable = self.output.stdout and (
            self.output.stdout.isatty()
            # Don't send escape codes if this is not a real TTY.
            # We create pseudo-ttys to get colored output, but don't want
            # any terminal queries to be sent
            and not getattr(self.output.stdout, "fake_tty", False)
        )
        if not self.queryable:
            self.cmd = ""

    def verify(self, data: str) -> Any | None:
        """Verify the response from the terminal."""
        return None

    async def _handle_response(self, event: KeyPressEvent) -> object:
        """Run when the terminal receives the response from the terminal.

        Args:
            event: The key press event received when the termina sends a response

        Returns:
            :py:obj:`NotImplemented`, so the application is not invalidated when a
            response from the terminal is received

        """
        log.debug("Got terminal response for '%s'", self.__class__.__name__)
        self.waiting = False
        new_value = self.verify(event.data)
        # Only fire the event if the value has changed
        if self._value != new_value:
            self._value = new_value
            self.event.fire()
        return NotImplemented

    def _cmd(self) -> str:
        """Return the query's command."""
        return self.cmd

    def send(self) -> None:
        """Send the terminal query command to the output."""
        if self.queryable and self.cmd and not self.waiting:
            cmd = self._cmd()
            log.debug(
                "Sending query %s for %s",
                cmd.__repr__(),
                self.__class__.__name__,
            )
            self.output.write_raw(cmd)
            self.output.flush()
            self.waiting = True

    @property
    def value(self) -> Any:
        """Return the last known value for the query.

        Returns:
            The last value received, or the default value.

        """
        return self._value or self.default


class Colors(TerminalQuery):
    """A terminal query to retrieve colours as hex codes."""

    _color_names = {
        "10": "fg",
        "11": "bg",
        "4;0": "ansiblack",
        "4;1": "ansired",
        "4;2": "ansigreen",
        "4;3": "ansiyellow",
        "4;4": "ansiblue",
        "4;5": "ansipurple",
        "4;6": "ansicyan",
        "4;7": "ansiwhite",
        "4;8": "ansirbightblack",
        "4;9": "ansirbightred",
        "4;10": "ansirbightgreen",
        "4;11": "ansirbightyellow",
        "4;12": "ansirbightblue",
        "4;13": "ansirbightpurple",
        "4;14": "ansirbightcyan",
        "4;15": "ansirbightwhite",
    }
    default = DEFAULT_COLORS
    cache = True
    cmd = ("\x1b]10;?\x1b\\" "\x1b]11;?\x1b\\") + "".join(
        f"\x1b]4;{i};?\x1b\\" for i in range(16)
    )
    pattern = re.compile(
        r"^\x1b\](?P<c>(\d+;)?\d+)+;rgb:"
        r"(?P<r>[0-9A-Fa-f]{2,4})\/"
        r"(?P<g>[0-9A-Fa-f]{2,4})\/"
        r"(?P<b>[0-9A-Fa-f]{2,4})"
        # Allow BEL or ST as terminator
        r"(\x1b\\\Z|\0x7)"
    )

    def _cmd(self) -> str:
        return tmuxify(self.cmd)

    def verify(self, data: str) -> dict[str, str]:
        """Verify the response contains a colour."""
        if match := self.pattern.match(data):
            if colors := match.groupdict():
                c = colors["c"]
                r, g, b = (
                    colors.get("r", "00"),
                    colors.get("g", "00"),
                    colors.get("b", "00"),
                )
                output = {
                    **self.value,
                    self._color_names.get(c, c): f"#{r[:2]}{g[:2]}{b[:2]}",
                }
                return output
        return self.value


class PixelDimensions(TerminalQuery):
    """A terminal query to check the terminal's dimensions in pixels."""

    default = (0, 0)
    cmd = "\x1b[14t"
    cache = True
    pattern = re.compile(r"^\x1b\[4;(?P<y>\d+);(?P<x>\d+)t\x1b\\")

    def verify(self, data: str) -> tuple[int, int] | None:
        """Verify the terminal responded with pixel dimensions."""
        if match := self.pattern.match(data):
            if values := match.groupdict():
                if (x := values.get("x")) is not None and (
                    y := values.get("y")
                ) is not None:
                    return int(x), int(y)
        return None


class KittyGraphicsStatus(TerminalQuery):
    """A terminal query to check for kitty graphics support."""

    default = False
    cache = True
    cmd = "\x1b_Gi=4294967295,s=1,v=1,a=q,t=d,f=24;AAAA\x1b\\"
    # Konsole responds with 'i=0' - I'll allow it
    pattern = re.compile(r"^\x1b_Gi=(4294967295|0);(?P<status>OK)\x1b\\\Z")

    def _cmd(self) -> str:
        """Hide the command in case the terminal does not support this sequence."""
        return "\x1b[s" + tmuxify(self.cmd) + "\x1b[u\x1b[2K"

    def verify(self, data: str) -> bool:
        """Verify the terminal response means kitty graphics are supported."""
        if match := self.pattern.match(data):
            if values := match.groupdict():
                if values.get("status") == "OK":
                    return True
        return False


class SixelGraphicsStatus(TerminalQuery):
    """A terminal query to check for sixel graphics support."""

    default = False
    cache = True
    cmd = "\x1b[c"
    pattern = re.compile(r"^\x1b\[\?(?:\d+;)*(?P<sixel>4)(?:;\d+)*c\Z")

    def _cmd(self) -> str:
        if self.config.tmux_graphics:
            return tmuxify(self.cmd)
        else:
            return self.cmd

    def verify(self, data: str) -> bool:
        """Verify the terminal response means sixel graphics are supported."""
        if match := self.pattern.match(data):
            if values := match.groupdict():
                if values.get("sixel"):
                    return True
        return False


class ItermGraphicsStatus(TerminalQuery):
    """A terminal query to check for iTerm graphics support."""

    default = False
    cache = True
    cmd = "\x1b[>q"
    pattern = re.compile(r"^\x1bP>\|(?P<term>[^\x1b]+)\x1b\\")

    def _cmd(self) -> str:
        if self.config.tmux_graphics:
            return tmuxify(self.cmd)
        else:
            return self.cmd

    def verify(self, data: str) -> bool:
        """Verify iterm graphics are supported by the terminal."""
        if match := self.pattern.match(data):
            if values := match.groupdict():
                if term := values.get("term"):
                    if (
                        term.startswith("WezTerm")
                        or term.startswith("Konsole")
                        or term.startswith("mlterm")
                    ):
                        return True
        return False


class DepthOfColor(TerminalQuery):
    """Determine the suspected color depth of the terminal."""

    # TODO - detect 24bit color support with escape sequence
    # "\x1b[48:2:1:2:3m\eP$qm\x1b\\"

    default = ColorDepth.DEPTH_24_BIT

    def __init__(self, output: Output, config: Config) -> None:
        """Detect the terminal's colour support based on environment variables."""
        super().__init__(output, config)
        self._value: ColorDepth | None = None
        if os.environ.get("NO_COLOR", "") or os.environ.get("TERM", "") == "dumb":
            self._value = ColorDepth.DEPTH_1_BIT
            return
        colorterm = os.environ.get("COLORTERM", "")
        if "truecolor" in colorterm or "24bit" in colorterm:
            self._value = ColorDepth.DEPTH_24_BIT
            return
        if "256" in os.environ.get("TERM", ""):
            self._value = ColorDepth.DEPTH_8_BIT
            return


class SgrPixelStatus(TerminalQuery):
    """A terminal query to check for Pixel SGR support."""

    default = False
    cache = True
    cmd = "\x1b[?1016h\x1b[?1016$p\x1b[?1016l"  # Enable, check, disable
    pattern = re.compile(r"^\x1b\[\?1016;(?P<Pm>\d)\$\Z")

    def verify(self, data: str) -> bool:
        """Verify the terminal response means SGR pixel-mode is supported."""
        if match := self.pattern.match(data):
            if values := match.groupdict():
                if values.get("Pm") in {"1", "3"}:
                    return True
        return False


class CsiUStatus(TerminalQuery):
    """A terminal query to check for CSI-u support."""

    default = False
    cache = True
    cmd = "\x1b[?u"
    pattern = re.compile(r"^\x1b\[\?\d+u")

    def verify(self, data: str) -> bool:
        """Verify the terminal responds."""
        if match := self.pattern.match(data):
            if match:
                return True
        return False


class TerminalInfo:
    """A class to gather and hold information about the terminal."""

    input: Input
    output: Output

    _queries: dict[type[TerminalQuery], TerminalQuery] = {}

    def __init__(self, input_: Input, output: Output, config: Config) -> None:
        """Instantiate the terminal information class."""
        self.input = input_
        self.output = output
        self.config = config

        self.colors = self.register(Colors)
        self.pixel_dimensions = self.register(PixelDimensions)
        self.sixel_graphics_status = self.register(SixelGraphicsStatus)
        self.kitty_graphics_status = self.register(KittyGraphicsStatus)
        self.iterm_graphics_status = self.register(ItermGraphicsStatus)
        self.depth_of_color = self.register(DepthOfColor)
        self.sgr_pixel_status = self.register(SgrPixelStatus)
        # self.csiu_status = self.register(CsiUStatus)

    def register(self, query: type[TerminalQuery]) -> TerminalQuery:
        """Instantiate and registers a query's response with the input parser."""
        # Create an instance of this query
        query_inst: TerminalQuery | None

        if (query_inst := self._queries.get(query)) is None:
            query_inst = query(self.output, config=self.config)
            self._queries[query] = query_inst

            # If the query expects a response from the terminal, we need to add a
            # key-binding for it and register it with the input parser
            if query.pattern:
                name = re.sub(r"(?<!^)(?=[A-Z])", "-", query.__name__).lower()
                title = name.replace("-", " ")

                # Add a "key" definition for this query
                key_name = f"{query.__name__}Response"
                key_code = f"<{name}-response>"
                # Do not register the same key multiple times
                if not hasattr(Keys, key_name):
                    extend_enum(Keys, key_name, key_code)
                key = getattr(Keys, key_name)

                # Register this key with the parser if supported
                if (parser := getattr(self.input, "vt100_parser", None)) and hasattr(
                    parser, "queries"
                ):
                    # Register the key
                    parser.queries[key] = query.pattern

                # Add a command for the query's key-binding
                add_cmd(name=name, title=title, hidden=True)(
                    query_inst._handle_response
                )
                # Add key-binding
                register_bindings({"euporie.core.app.BaseApp": {name: key}})

        return query_inst

    def send_all(self) -> None:
        """Send the command for all queries."""
        # Ensure line wrapping is off before sending queries
        self.output.disable_autowrap()
        for query in self._queries.values():
            query.send()

    def _tiocgwnsz(self) -> tuple[int, int, int, int]:
        """Get the size and pixel dimensions of the terminal with `termios`."""
        output = array.array("H", [0, 0, 0, 0])
        if _have_termios_tty_fcntl():
            import fcntl
            import termios

            try:
                fcntl.ioctl(1, termios.TIOCGWINSZ, output)
            except Exception:  # noqa S110
                pass
        rows, cols, xpixels, ypixels = output
        return rows, cols, xpixels, ypixels

    @property
    def terminal_size_px(self) -> tuple[int, int]:
        """Get the pixel dimensions of the terminal."""
        # Prefer using escape codes as this works over SSH
        px, py = self.pixel_dimensions.value
        if px == 0:
            # If unsuccessful, try requesting info with tiocgwnsz
            _rows, _cols, px, py = self._tiocgwnsz()
        return px, py

    @property
    def cell_size_px(self) -> tuple[int, int]:
        """Get the pixel size of a single terminal cell."""
        px, py = self.terminal_size_px
        rows, cols = self.output.get_size()
        # If we can't get the pixel size, just guess wildly
        return px // cols or 10, py // rows or 20


def edit_in_editor(filename: str, line_number: int = 0) -> None:
    """Suspend the current app and edit a file in an external editor."""
    log.debug(filename)

    def _open_file_in_editor(filename: str) -> None:
        """Call editor executable."""
        # If the 'VISUAL' or 'EDITOR' environment variable has been set, use that.
        # Otherwise, fall back to the first available editor that we can find.
        for editor in [
            os.environ.get("VISUAL"),
            os.environ.get("EDITOR"),
            "editor",
            "micro",
            "nano",
            "pico",
            "vi",
            "emacs",
        ]:
            if editor:
                try:
                    # Use 'shlex.split()' because $VISUAL can contain spaces and quotes
                    subprocess.call(shlex.split(editor) + [filename])
                    return
                except OSError:
                    # Executable does not exist, try the next one.
                    pass

    async def run() -> None:
        # Open in editor
        # (We need to use `run_in_terminal`, because not all editors go to
        # the alternate screen buffer, and some could influence the cursor
        # position)
        await run_in_terminal(lambda: _open_file_in_editor(filename), in_executor=True)

    get_app().create_background_task(run())



"""Define the state of the micro editing mode."""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from prompt_toolkit.key_binding.key_processor import KeyPress


class MicroInputMode(str, Enum):
    """Enum to define edit mode state types."""

    value: str
    INSERT = "insert"
    REPLACE = "replace"


class MicroState:
    """Mutable class to hold Micro specific state."""

    def __init__(self) -> None:
        """Initiate the editing mode state."""
        self.macro: list[KeyPress] | None = []
        self.current_recording: list[KeyPress] | None = None
        self.input_mode: MicroInputMode = MicroInputMode.INSERT

    def reset(self) -> None:
        """Reet the editing mode state."""
        self.input_mode = MicroInputMode.INSERT
        self.current_recording = None

    @property
    def is_recording(self) -> bool:
        """Tell whether we are recording a macro."""
        return self.current_recording is not None

    def start_macro(self) -> None:
        """Start recording a macro."""
        self.current_recording = []

    def end_macro(self) -> None:
        """End recording a macro."""
        self.macro = self.current_recording
        self.current_recording = None



"""Define enums."""

from enum import Enum


class TabMode(Enum):
    """Define how multiple tabs are displayed."""

    STACK = "stack"
    TILE_HORIZONTALLY = "tile_horizontally"
    TILE_VERTICALLY = "tile_vertically"



"""Contain functions which convert data to PIL format."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from euporie.core.convert.core import register
from euporie.core.convert.utils import have_modules

if TYPE_CHECKING:
    from pathlib import Path

    from PIL.Image import Image as PilImage

log = logging.getLogger(__name__)


def set_background(image: PilImage, bg_color: str | None = None) -> PilImage:
    """Remove the alpha channel from an image and set the background colour."""
    from PIL import Image

    if image.mode in ("RGBA", "LA") or (
        image.mode == "P" and "transparency" in image.info
    ):
        alpha = image.convert("RGBA").getchannel("A")
        bg = Image.new("RGBA", image.size, bg_color or "#000")
        bg.paste(image, mask=alpha)
        image = bg
    image = image.convert("P", palette=Image.Palette.ADAPTIVE, colors=16).convert(
        "RGB", palette=Image.Palette.ADAPTIVE, colors=16
    )
    return image


@register(
    from_=("png", "jpeg", "gif"),
    to="pil",
    filter_=have_modules("PIL"),
)
async def png_to_pil_py(
    data: bytes,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    path: Path | None = None,
    initial_format: str = "",
) -> PilImage:
    """Convert PNG to a pillow image using :py:mod:`PIL`."""
    import io

    from PIL import Image

    try:
        image = Image.open(io.BytesIO(data))
        image.load()
    except OSError:
        log.error("Could not load image.")
        return Image.new(mode="P", size=(1, 1))
    else:
        return image



"""Utilitie for manipulating formatted text."""

from __future__ import annotations

import re
from enum import Enum
from typing import Iterable, cast

from prompt_toolkit.formatted_text.base import OneStyleAndTextTuple, StyleAndTextTuples
from prompt_toolkit.formatted_text.utils import (
    fragment_list_to_text,
    split_lines,
    to_plain_text,
)
from prompt_toolkit.layout.utils import explode_text_fragments
from prompt_toolkit.utils import get_cwidth
from pygments.lexers import get_lexer_by_name
from pygments.util import ClassNotFound

from euporie.core.border import GridStyle, ThinGrid
from euporie.core.data_structures import DiBool, DiInt, DiStr

_ZERO_WIDTH_FRAGMENTS = ("[ZeroWidthEscape]", "[ReverseOverwrite]")


class FormattedTextAlign(Enum):
    """Alignment of formatted text."""

    LEFT = "left"
    RIGHT = "right"
    CENTER = "center"


class FormattedTextVerticalAlign(Enum):
    """Vertical alignment of formatted text."""

    TOP = "top"
    MIDDLE = "middle"
    BOTTOM = "bottom"


def fragment_list_width(fragments: StyleAndTextTuples) -> int:
    """Return the character width of this text fragment list.

    Takes double width characters into account, and ignore special fragments:
    * ZeroWidthEscape
    * ReverseOverwrite

    Args:
        fragments: List of ``(style_str, text)`` or
            ``(style_str, text, mouse_handler)`` tuples.

    Returns:
        The width of the fragment list
    """
    return sum(
        get_cwidth(c)
        for item in (
            frag
            for frag in fragments
            if not any(x in frag[0] for x in _ZERO_WIDTH_FRAGMENTS)
        )
        for c in item[1]
    )


def last_line_length(ft: StyleAndTextTuples, rows: int = 1) -> int:
    """Calculate the length of the last line in formatted text."""
    lines: list[StyleAndTextTuples] = [[]]
    for style, text, *_ in reversed(ft):
        parts = text.split("\n")
        lines[-1].append((style, parts[-1]))
        if parts:
            for part in parts[-2::-1]:
                if len(lines) == rows:
                    break
                lines.append([])
                lines[-1].append((style, part))
            else:
                continue
            break
    return max(fragment_list_width(line) for line in lines)


def max_line_width(ft: StyleAndTextTuples) -> int:
    """Calculate the length of the longest line in formatted text."""
    return max(fragment_list_width(line) for line in split_lines(ft))


def last_char(ft: StyleAndTextTuples) -> str | None:
    """Retrieve the last character of formatted text."""
    for frag in reversed(ft):
        text = frag[1]
        for c in reversed(text):
            if c:
                return c
    return None


def fragment_list_to_words(
    fragments: StyleAndTextTuples, sep: str = " "
) -> Iterable[StyleAndTextTuples]:
    """Split formatted text into a list of word fragments which form words."""
    word: StyleAndTextTuples = []
    for style, string, *rest in fragments:
        parts = re.split(r"(?<=[\s\-])", string)
        if len(parts) == 1:
            word.append(cast("OneStyleAndTextTuple", (style, parts[0], *rest)))
        else:
            for part in parts[:-1]:
                word.append(cast("OneStyleAndTextTuple", (style, part, *rest)))
                yield word[:]
                word.clear()
            word.append(cast("OneStyleAndTextTuple", (style, parts[-1], *rest)))
    if word:
        yield word


def apply_style(ft: StyleAndTextTuples, style: str) -> StyleAndTextTuples:
    """Apply a style to formatted text."""
    return [
        (
            f"{fragment_style} {style}"
            if "[ZeroWidthEscape]" not in fragment_style
            else fragment_style,
            text,
        )
        for (fragment_style, text, *_) in ft
    ]


def strip(
    ft: StyleAndTextTuples,
    left: bool = True,
    right: bool = True,
    chars: str | None = None,
    only_unstyled: bool = False,
) -> StyleAndTextTuples:
    """Strip whitespace (or a given character) from the ends of formatted text.

    Args:
        ft: The formatted text to strip
        left: If :py:const:`True`, strip from the left side of the input
        right: If :py:const:`True`, strip from the right side of the input
        chars: The character to strip. If :py:const:`None`, strips whitespace
        only_unstyled: If :py:const:`True`, only strip unstyled fragments

    Returns:
        The stripped formatted text

    """
    result = ft[:]
    for toggle, index, strip_func in [(left, 0, str.lstrip), (right, -1, str.rstrip)]:
        if result and toggle:
            text = strip_func(result[index][1], chars)
            while result and not text:
                del result[index]
                if not result:
                    break
                text = strip_func(result[index][1], chars)
            if result and "[ZeroWidthEscape]" not in result[index][0]:
                result[index] = (result[index][0], text)
    return result


def strip_one_trailing_newline(ft: StyleAndTextTuples) -> StyleAndTextTuples:
    """Remove up to one trailing new-line character from formatted text."""
    for i in range(len(ft) - 1, -1, -1):
        frag = ft[i]
        if not frag[1]:
            continue
        if frag[1] == "\n":
            del ft[i]
        elif frag[1].endswith("\n"):
            ft[i] = (frag[0], frag[1][:-1])
        break
    return ft


def truncate(
    ft: StyleAndTextTuples,
    width: int,
    style: str = "",
    placeholder: str = "…",
    ignore_whitespace: bool = False,
) -> StyleAndTextTuples:
    """Truncate all lines at a given length.

    Args:
        ft: The formatted text to truncate
        width: The width at which to truncate the text
        style: The style to apply to the truncation placeholder. The style of the
            truncated text will be used if not provided
        placeholder: The string that will appear at the end of a truncated line
        ignore_whitespace: Do not use placeholder when truncating whitespace

    Returns:
        The truncated formatted text

    """
    lines = split_lines(ft)
    if max(fragment_list_width(line) for line in lines) <= width:
        return ft
    result: StyleAndTextTuples = []
    phw = sum(get_cwidth(c) for c in placeholder)
    for line in split_lines(ft):
        used_width = 0
        for i, item in enumerate(line):
            fragment_width = sum(
                get_cwidth(c)
                for c in item[1]
                if not any(x in item[0] for x in _ZERO_WIDTH_FRAGMENTS)
            )
            if (
                used_width + fragment_width > width - phw
                # Do not truncate if we have the exact width
                and used_width + fragment_width != width
            ):
                remaining_width = width - used_width - fragment_width - phw
                result.append((item[0], item[1][:remaining_width]))
                if ignore_whitespace and not fragment_list_to_text(line[i:]).strip():
                    result.append(
                        (item[0], item[1][remaining_width : remaining_width + 1])
                    )
                else:
                    result.append((style or item[0], placeholder))
                break
            else:
                result.append(item)
                used_width += fragment_width
        result.append(("", "\n"))
    result.pop()
    return result


def substring(
    ft: StyleAndTextTuples, start: int | None = None, end: int | None = None
) -> StyleAndTextTuples:
    """Extract a substring from formatted text."""
    output: StyleAndTextTuples = []
    if start is None:
        start = 0
    if end is None or start < 0 or end < 0:
        width = fragment_list_width(ft)
        if end is None:
            end = width
        if start < 0:
            start = width + start
        if end < 0:
            end = width + end
    x = 0
    for style, text, *extra in ft:
        if any(x in style for x in _ZERO_WIDTH_FRAGMENTS):
            frag_len = 0
        else:
            frag_len = sum(get_cwidth(c) for c in text)
        if start <= x + frag_len <= end + frag_len:
            if (text := text[max(0, start - x) : end - x]) or style:
                output.append(cast("OneStyleAndTextTuple", (style, text, *extra)))
        x += frag_len
    return output


def wrap(
    ft: StyleAndTextTuples,
    width: int,
    style: str = "",
    placeholder: str = "…",
    left: int = 0,
    truncate_long_words: bool = True,
    strip_trailing_ws: bool = False,
) -> StyleAndTextTuples:
    """Wrap formatted text at a given width.

    If words are longer than the given line they will be truncated

    Args:
        ft: The formatted text to wrap
        width: The width at which to wrap the text
        style: The style to apply to the truncation placeholder
        placeholder: The string that will appear at the end of a truncated line
        left: The starting position within the first line
        truncate_long_words: If :const:`True` words longer than a line will be
            truncated
        strip_trailing_ws: If :const:`True`, trailing whitespace will be removed from
            the ends of lines

    Returns:
        The wrapped formatted text
    """
    result: StyleAndTextTuples = []
    lines = list(split_lines(ft))
    output_line = 0
    for i, line in enumerate(lines):
        if fragment_list_width(line) <= width - left:
            result += line
            if i < len(lines) - 1:
                result.append(("", "\n"))
            output_line += 1
            left = 0
        else:
            for word in fragment_list_to_words(line):
                # Skip empty fragments
                # if item[1] == "":
                # continue

                fragment_width = fragment_list_width(word)

                # Start a new line - we are at the end of the current output line
                if left + fragment_width > width and left > 0:
                    # Add as much trailing whitespace as we can
                    if not strip_trailing_ws and (
                        trailing_ws := (not to_plain_text(word).strip())
                    ):
                        result.extend(substring(word, end=width - left))

                    # Remove trailing whitespace
                    if strip_trailing_ws:
                        result = strip(result, left=False)

                    # Start new line
                    result.append(("", "\n"))
                    output_line += 1
                    left = 0

                    # If we added trailing whitespace, process the next fragment
                    if not strip_trailing_ws and trailing_ws:
                        continue

                # Strip left-hand whitespace from a word at the start of a line
                # if output_line != 0 and left == 0:
                if left == 0:
                    word = strip(word, right=False)
                # Truncate words longer than a line
                if truncate_long_words and left == 0 and fragment_width > width - left:
                    result += truncate(word, width - left, style, placeholder)
                    left += fragment_width
                # Otherwise just add the word to the line
                else:
                    result.extend(word)
                    left += fragment_width

    if strip_trailing_ws:
        result = strip(result, left=False)

    return result


def align(
    ft: StyleAndTextTuples,
    how: FormattedTextAlign = FormattedTextAlign.LEFT,
    width: int | None = None,
    style: str = "",
    placeholder: str = "…",
    ignore_whitespace: bool = False,
) -> StyleAndTextTuples:
    """Align formatted text at a given width.

    Args:
        how: The alignment direction
        ft: The formatted text to strip
        width: The width to which the output should be padded. If :py:const:`None`, the
            length of the longest line is used
        style: The style to apply to the padding
        placeholder: The string that will appear at the end of a truncated line
        ignore_whitespace: If True, whitespace will be ignored

    Returns:
        The aligned formatted text

    """
    style = f"{style} nounderline"
    lines = split_lines(ft)
    if width is None:
        lines = [strip(line) if ignore_whitespace else line for line in split_lines(ft)]
        width = max(fragment_list_width(line) for line in lines)
    result: StyleAndTextTuples = []

    for line in lines:
        line_width = fragment_list_width(line)
        # Truncate the line if it is too long
        if line_width > width:
            result += truncate(line, width, style, placeholder, ignore_whitespace)
        else:
            pad_left = pad_right = 0
            if how == FormattedTextAlign.CENTER:
                pad_left = (width - line_width) // 2
                pad_right = width - line_width - pad_left
            elif how == FormattedTextAlign.LEFT:
                pad_right = width - line_width
            elif how == FormattedTextAlign.RIGHT:
                pad_left = width - line_width
            if pad_left:
                result.append((style, " " * pad_left))
            result.extend(line)
            if pad_right:
                result.append((style, " " * pad_right))
        result.append((style, "\n"))
    result.pop()
    return result


def valign(
    ft: StyleAndTextTuples,
    how: FormattedTextVerticalAlign = FormattedTextVerticalAlign.MIDDLE,
    height: int | None = None,
    style: str = "",
) -> StyleAndTextTuples:
    """Align formatted text vertically."""
    if height is None:
        return ft
    lines = list(split_lines(ft))
    width = max(fragment_list_width(line) for line in lines)
    remaining = height - len(lines)
    if how == FormattedTextVerticalAlign.TOP:
        above = 0
        below = remaining
    elif how == FormattedTextVerticalAlign.MIDDLE:
        above = remaining // 2
        below = remaining - above
    else:
        above = remaining
        below = 0
    return [
        (style, ((" " * width) + "\n") * above),
        *ft,
        (style, ("\n" + (" " * width)) * below),
    ]


def join_lines(fragments: list[StyleAndTextTuples]) -> StyleAndTextTuples:
    """Join a list of lines of formatted text."""
    ft = []
    if fragments:
        for line in fragments[:-1]:
            ft.extend(line)
            ft.append(("", "\n"))
        lines = list(split_lines(fragments[-1]))
        for line in lines[:-1]:
            ft.extend(line)
            ft.append(("", "\n"))
        ft.extend(lines[-1])
    return ft


def pad(
    ft: StyleAndTextTuples,
    width: int | None = None,
    char: str = " ",
    style: str = "",
) -> StyleAndTextTuples:
    """Fill space at the end of lines."""
    if width is None:
        width = max_line_width(ft)
    filled_output = []
    for line in split_lines(ft):
        if (remaining := (width - fragment_list_width(line))) > 0:
            line.append((style + " nounderline", (char * remaining)))
        filled_output.append(line)
    return join_lines(filled_output)


def paste(
    ft_top: StyleAndTextTuples,
    ft_bottom: StyleAndTextTuples,
    row: int = 0,
    col: int = 0,
    transparent: bool = False,
) -> StyleAndTextTuples:
    """Pate formatted text on top of other formatted text."""
    ft: StyleAndTextTuples = []
    top_lines = dict(enumerate(split_lines(ft_top), start=row))
    for y, line_b in enumerate(split_lines(ft_bottom)):
        if y in top_lines:
            line_t = top_lines[y]
            line_t_width = fragment_list_width(line_t)
            ft += substring(line_b, 0, col)
            if transparent:
                chars_t = explode_text_fragments(line_t)
                chars_b = explode_text_fragments(
                    substring(line_b, col, col + line_t_width)
                )
                for char_t, char_b in zip(chars_t, chars_b):
                    if char_t[0] == "" and char_t[1] == " ":
                        ft.append(char_b)
                    else:
                        ft.append(char_t)
            else:
                ft += line_t
            ft += substring(line_b, col + line_t_width)
        else:
            ft += line_b
        ft.append(("", "\n"))

    if ft:
        ft.pop()

    return ft


def concat(
    ft_a: StyleAndTextTuples,
    ft_b: StyleAndTextTuples,
    baseline_a: int = 0,
    baseline_b: int = 0,
    style: str = "",
) -> tuple[StyleAndTextTuples, int]:
    """Concatenate two blocks of formatted text, aligning at a given baseline.

    Args:
        ft_a: The first block of formatted text to combine
        ft_b: The second block of formatted text to combine
        baseline_a: The row to use to align the first block of formatted text with the
            second, counted in lines down from the top of the block
        baseline_b: The row to use to align the second block of formatted text with the
            second, counted in lines down from the top of the block
        style: The style to use for any extra lines added

    Returns:
        A tuple containing the combined formatted text and the new baseline position
    """
    rows_a = len(list(split_lines(ft_a))) - 1
    rows_b = len(list(split_lines(ft_b))) - 1
    cols_a = max_line_width(ft_a)

    lines_above = max(baseline_a, baseline_b) - min(baseline_a, baseline_b)
    lines_below = max(rows_a - baseline_a, rows_b - baseline_b) - min(
        rows_a - baseline_a,
        rows_b - baseline_b,
    )

    if baseline_a < baseline_b:
        ft_a = [("", lines_above * "\n"), *ft_a]
    elif baseline_a > baseline_b:
        ft_b = [("", lines_above * "\n"), *ft_b]

    if rows_a - baseline_a < rows_b - baseline_b:
        ft_a = [*ft_a, ("", lines_below * "\n")]
    elif rows_a - baseline_a > rows_b - baseline_b:
        ft_b = [*ft_b, ("", lines_below * "\n")]

    ft = paste(
        ft_b,
        pad(ft_a, style=style),
        row=0,
        col=cols_a,
    )

    new_baseline = max(baseline_a, baseline_b)

    return ft, new_baseline


def indent(
    ft: StyleAndTextTuples,
    margin: str = " ",
    style: str = "",
    skip_first: bool = False,
) -> StyleAndTextTuples:
    """Indent formatted text with a given margin.

    Args:
        ft: The formatted text to strip
        margin: The margin string to add
        style: The style to apply to the margin
        skip_first: If :py:const:`True`, the first line is skipped

    Returns:
        The indented formatted text

    """
    result: StyleAndTextTuples = []
    for i, line in enumerate(split_lines(ft)):
        if not (i == 0 and skip_first):
            result.append((style, margin))
        result += line
        result.append(("", "\n"))
    result.pop()
    return result


def add_border(
    ft: StyleAndTextTuples,
    width: int | None = None,
    style: str = "",
    border_grid: GridStyle = ThinGrid,
    border_visibility: DiBool | bool = True,
    border_style: DiStr | str = "",
    padding: DiInt | int = 0,
    padding_style: DiStr | str = "",
) -> StyleAndTextTuples:
    """Add a border around formatted text.

    Args:
        ft: The formatted text to enclose with a border
        width: The target width including the border and padding
        style: The style to apply to the content background
        border_grid: The grid style to use for the border
        border_visibility: Determines which edges should receive a border
        border_style: The style to apply to the border
        padding: The width of spacing to apply between the content and the border
        padding_style: The style to apply to the border

    Returns:
        The indented formatted text

    """
    if isinstance(border_visibility, bool):
        border_visibility = DiBool.from_value(border_visibility)
    if isinstance(padding, int):
        padding = DiInt.from_value(padding)
    if isinstance(border_style, str):
        border_style = DiStr.from_value(border_style)
    if isinstance(padding_style, str):
        padding_style = DiStr.from_value(padding_style)

    edge_width = (
        border_visibility.left + border_visibility.right + padding.left + padding.right
    )
    if width is None:
        inner_width = max_line_width(ft)
        width = inner_width + edge_width
    else:
        inner_width = width - edge_width

    # Ensure all lines are the same length
    if ft:
        ft = align(
            ft,
            FormattedTextAlign.LEFT,
            width=inner_width,
            style=style,
        )

    output: list[StyleAndTextTuples] = []

    base_style = f"{style} nounderline"

    # Add top row
    if border_visibility.top:
        new_line: StyleAndTextTuples = []
        if border_visibility.left:
            new_line.append(
                (
                    f"{base_style} class:border,left,top {border_style.top} {border_style.left}",
                    border_grid.TOP_LEFT,
                )
            )
        new_line.append(
            (
                f"{base_style} class:border,top {border_style.top}",
                border_grid.TOP_MID * (inner_width + padding.left + padding.right),
            )
        )
        if border_visibility.right:
            new_line.append(
                (
                    f"{base_style} class:border,right,top {border_style.top} {border_style.right}",
                    border_grid.TOP_RIGHT,
                )
            )
        output.append(new_line)
    # Add top padding
    for _ in range(padding.top or 0):
        new_line = []
        if border_visibility.left:
            new_line.append(
                (
                    f"{base_style} class:border,left {border_style.left}",
                    border_grid.MID_LEFT,
                )
            )
        if padding.top:
            new_line.append(
                (
                    f"{base_style} class:padding,top {padding_style.top}",
                    " " * (inner_width + padding.left + padding.right),
                )
            )
        if border_visibility.right:
            new_line.append(
                (
                    f"{base_style} class:border,right {border_style.right}",
                    border_grid.MID_RIGHT,
                )
            )
        output.append(new_line)
    # Add contents with left & right padding
    if ft:
        for line in split_lines(ft):
            new_line = []
            if border_visibility.left:
                new_line.append(
                    (
                        f"{base_style} class:border,left {border_style.left}",
                        border_grid.MID_LEFT,
                    )
                )
            if padding.left:
                new_line.append(
                    (
                        f"{base_style} class:padding,left {padding_style.left}",
                        " " * (padding.left or 0),
                    )
                )

            new_line.extend(line)
            if padding.right:
                new_line.append(
                    (
                        f"{base_style} class:padding,right {padding_style.right}",
                        " " * (padding.right or 0),
                    )
                )
            if border_visibility.right:
                new_line.append(
                    (
                        f"{base_style} class:border,right {border_style.right}",
                        border_grid.MID_RIGHT,
                    )
                )
            output.append(new_line)
    # Add bottom padding
    for _ in range(int(padding.bottom) or 0):
        new_line = []
        if border_visibility.left:
            new_line.append(
                (
                    f"{base_style} class:border,left {border_style.left}",
                    border_grid.MID_LEFT,
                )
            )
        if padding.bottom:
            new_line.append(
                (
                    f"{base_style} class:padding,bottom {padding_style.bottom}",
                    " " * (inner_width + padding.left + padding.right),
                )
            )
        if border_visibility.right:
            new_line.append(
                (
                    f"{base_style} class:border,right {border_style.right}",
                    border_grid.MID_RIGHT,
                )
            )
        output.append(new_line)
    # Add bottom row
    if border_visibility.bottom:
        new_line = []
        if border_visibility.left:
            new_line.append(
                (
                    f"{base_style} class:border,left,bottom {border_style.bottom} {border_style.left}",
                    border_grid.BOTTOM_LEFT,
                )
            )
        new_line.append(
            (
                f"{base_style} class:border,bottom {border_style.bottom}",
                border_grid.BOTTOM_MID * (inner_width + padding.left + padding.right),
            )
        )
        if border_visibility.right:
            new_line.append(
                (
                    f"{base_style} class:border,right,bottom {border_style.bottom} {border_style.right}",
                    border_grid.BOTTOM_RIGHT,
                )
            )

        output.append(new_line)

    return join_lines(output)


def lex(ft: StyleAndTextTuples, lexer_name: str) -> StyleAndTextTuples:
    """Format formatted text using a named :py:mod:`pygments` lexer."""
    from prompt_toolkit.lexers.pygments import _token_cache

    try:
        lexer = get_lexer_by_name(lexer_name)
    except ClassNotFound:
        return ft
    else:
        output: StyleAndTextTuples = []
        for style, text, *_rest in ft:
            for _, t, v in lexer.get_tokens_unprocessed(text):
                output.append((f"{style} {_token_cache[t]}", v))
        return output


def apply_reverse_overwrites(ft: StyleAndTextTuples) -> StyleAndTextTuples:
    """Write fragments tagged with "[ReverseOverwrite]" over text to their left."""

    def _apply_overwrites(
        overwrites: StyleAndTextTuples, transformed_line: StyleAndTextTuples
    ) -> tuple[StyleAndTextTuples, StyleAndTextTuples]:
        """Pate `overwrites` over the end of `transformed_line`."""
        # Remove the ``[ReverseOverwrite]`` from the overwrite fragments
        top = cast(
            StyleAndTextTuples,
            [(x[0].replace("[ReverseOverwrite]", ""), *x[1:]) for x in overwrites],
        )
        top_width = fragment_list_width(top)
        if fragment_list_width(transformed_line) >= top_width:
            # Replace the end of the line with the overwrite-fragments if the overwrite
            # -fragments are shorter than the line
            transformed_line = [*substring(transformed_line, 0, -top_width), *top]
        else:
            # Otherwise, replace the whole line with the overwrites
            transformed_line = overwrites[:]
        overwrites.clear()
        return overwrites, transformed_line

    transformed_lines = []
    for untransformed_line in split_lines(ft):
        overwrites = []
        transformed_line: StyleAndTextTuples = []
        for frag in untransformed_line:
            # Collect overwrite fragments
            if "[ReverseOverwrite]" in frag[0]:
                overwrites.append(frag)
                continue
            # If we have any overwrite-fragments, apply them on top of the current line
            if overwrites:
                overwrites, transformed_line = _apply_overwrites(
                    overwrites, transformed_line
                )
            # Collect non-overwrite fragments
            transformed_line.append(frag)
        # Add this list to
        transformed_lines.append(transformed_line)
    return join_lines(transformed_lines)



"""Define widget for defining layouts."""

from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from functools import partial
from typing import TYPE_CHECKING, NamedTuple, cast

from prompt_toolkit.application.current import get_app
from prompt_toolkit.cache import SimpleCache
from prompt_toolkit.filters import Condition, to_filter
from prompt_toolkit.formatted_text.base import to_formatted_text
from prompt_toolkit.formatted_text.utils import fragment_list_width
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    DynamicContainer,
    HSplit,
    VSplit,
    Window,
    to_container,
)
from prompt_toolkit.layout.controls import (
    FormattedTextControl,
    GetLinePrefixCallable,
    UIContent,
    UIControl,
)
from prompt_toolkit.mouse_events import MouseButton, MouseEventType
from prompt_toolkit.utils import Event
from prompt_toolkit.widgets import Box

from euporie.core.border import OutsetGrid
from euporie.core.data_structures import DiBool
from euporie.core.ft.utils import truncate
from euporie.core.widgets.decor import Border

if TYPE_CHECKING:
    from typing import Any, Callable, Sequence

    from prompt_toolkit.filters import FilterOrBool
    from prompt_toolkit.formatted_text.base import AnyFormattedText, StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.layout.containers import AnyContainer, Container, _Split
    from prompt_toolkit.layout.dimension import AnyDimension
    from prompt_toolkit.mouse_events import MouseEvent

    from euporie.core.border import GridStyle

log = logging.getLogger(__name__)


class ConditionalSplit:
    """A split container where the orientation depends on a filter."""

    def __init__(self, vertical: FilterOrBool, *args: Any, **kwargs: Any) -> None:
        """Create a new conditional split container.

        Args:
            vertical: A filter which determines if the container should be displayed vertically
            args: Positional arguments to pass to the split container
            kwargs: Key-word arguments to pass to the split container

        """
        self.vertical = to_filter(vertical)
        self.args = args
        self.kwargs = kwargs
        self._cache: SimpleCache = SimpleCache(maxsize=2)

    def load_container(self, vertical: bool) -> _Split:
        """Load the container."""
        if vertical:
            return HSplit(*self.args, **self.kwargs)
        else:
            return VSplit(*self.args, **self.kwargs)

    def container(self) -> _Split:
        """Return the container for the current orientation."""
        vertical = self.vertical()
        return self._cache.get(vertical, partial(self.load_container, vertical))

    def __pt_container__(self) -> AnyContainer:
        """Return a dymanic container."""
        return DynamicContainer(self.container)


class ReferencedSplit:
    """A split container which maintains a reference to it's children."""

    def __init__(
        self,
        split: type[_Split],
        children: Sequence[AnyContainer],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Create a new instance of the split container.

        Args:
            split: The split container class (:class:`HSplit` or :class:`VSplit`)
            children: A list of child containers
            args: Positional arguments to pass to the split container
            kwargs: Key-word arguments to pass to the split container

        """
        self.container = split([], *args, **kwargs)
        self.children = list(children)

    @property
    def children(self) -> list[AnyContainer]:
        """Convert the referenced children to containers."""
        return self._children

    @children.setter
    def children(self, children: list[AnyContainer]) -> None:
        """Set the containers children."""
        self._children = children
        self.container.children = [to_container(x) for x in self._children]

    def __pt_container__(self) -> Container:
        """Return the child container."""
        return self.container


class TabBarTab(NamedTuple):
    """A named tuple represting a tab and it's callbacks."""

    title: AnyFormattedText
    on_activate: Callable
    on_deactivate: Callable | None = None
    on_close: Callable | None = None


class TabBarControl(UIControl):
    """A control which shows a tab bar."""

    char_bottom = "▁"
    char_left = "▏"
    char_right = "▕"
    char_top = "▁"
    char_close = "✖"

    def __init__(
        self,
        tabs: Sequence[TabBarTab] | Callable[[], Sequence[TabBarTab]],
        active: int | Callable[[], int],
        spacing: int = 1,
        closeable: bool = False,
        max_title_width: int = 30,
    ) -> None:
        """Create a new tab bar instance.

        Args:
            tabs: A list to tuples describing the tab title and the callback to run
                when the tab is activated.
            active: The index of the currently active tab
            spacing: The number of characters between the tabs
            closeable: Whether to show close buttons the the tabs
            max_title_width: The maximum width of the title to display

        """
        self._tabs = tabs
        self.spacing = spacing
        self.closeable = closeable
        self.max_title_width = max_title_width
        self._active = active

        self.mouse_handlers: dict[int, Callable[..., Any] | None] = {}

        self._title_cache: SimpleCache = SimpleCache(maxsize=1)
        self._content_cache: SimpleCache = SimpleCache(maxsize=50)

    @property
    def tabs(self) -> list[TabBarTab]:
        """Return the tab-bar's tabs."""
        if callable(self._tabs):
            return list(self._tabs())
        else:
            return list(self._tabs)

    @tabs.setter
    def tabs(self, tabs: Sequence[TabBarTab]) -> None:
        """Set the tab bar's current tabs."""
        self._tabs = tabs

    @property
    def active(self) -> int:
        """Return the index of the active tab."""
        if callable(self._active):
            return self._active()
        else:
            return self._active

    @active.setter
    def active(self, active: int | Callable[[], int]) -> None:
        """Set the currently active tab."""
        self._active = active

    def preferred_width(self, max_available_width: int) -> int | None:
        """Return the preferred width of the tab-bar control, the maximum available."""
        return max_available_width

    def preferred_height(
        self,
        width: int,
        max_available_height: int,
        wrap_lines: bool,
        get_line_prefix: GetLinePrefixCallable | None,
    ) -> int | None:
        """Return the preferred height of the tab-bar control (2 rows)."""
        return 2

    def is_focusable(self) -> bool:
        """Tell whether this user control is focusable."""
        return True

    def create_content(self, width: int, height: int) -> UIContent:
        """Generate the formatted text fragments which make the controls output."""

        def get_content() -> UIContent:
            fragment_lines = self.render(width)

            return UIContent(
                get_line=lambda i: fragment_lines[i],
                line_count=len(fragment_lines),
                show_cursor=False,
            )

        key = (hash(tuple(self.tabs)), width, self.closeable, self.active)
        return self._content_cache.get(key, get_content)

    def render(self, width: int) -> list[StyleAndTextTuples]:
        """Render the tab-bar as linest of formatted text."""
        top_line: StyleAndTextTuples = []
        tab_line: StyleAndTextTuples = []
        i = 0

        # Initial spacing
        for _ in range(self.spacing):
            top_line += [("", " ")]
            tab_line += [("class:border,bottom", self.char_bottom)]
        i += self.spacing

        for j, tab in enumerate(self.tabs):
            title_ft = truncate(to_formatted_text(tab.title), self.max_title_width)
            title_width = fragment_list_width(title_ft)
            style = "class:active" if self.active == j else "class:inactive"

            # Add top edge over title
            top_line += [
                (f"{style} class:tab,border,top", self.char_top * (title_width + 2))
            ]

            # Left edge
            tab_line += [(f"{style} class:tab,border,left", self.char_left)]
            self.mouse_handlers[i] = tab.on_activate
            i += 1

            # Title
            tab_line += [
                (f"{style} class:tab,title {frag_style}", text)
                for frag_style, text, *_ in title_ft
            ]
            for _ in range(title_width):
                self.mouse_handlers[i] = tab.on_activate
                i += 1

            # Close button
            if self.closeable:
                top_line += [(f"{style} class:tab,border,top", self.char_top * 2)]
                self.mouse_handlers[i] = tab.on_activate
                i += 1
                tab_line += [
                    (f"{style} class:tab", " "),
                    (f"{style} class:tab,close", self.char_close),
                ]
                self.mouse_handlers[i] = tab.on_close
                i += 1

            # Right edge
            tab_line += [(f"{style} class:tab,border,right", self.char_right)]
            self.mouse_handlers[i] = tab.on_activate
            i += 1

            # Spacing
            for _ in range(self.spacing):
                top_line += [("", " ")]
                tab_line += [("class:border,bottom", self.char_bottom)]
                i += 1

        # Add border to fill width
        tab_line += [
            (
                "class:border,bottom",
                self.char_bottom * (width - fragment_list_width(tab_line)),
            )
        ]

        result = [top_line, tab_line]
        return result

    def mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Handle mouse events."""
        row = mouse_event.position.y
        col = mouse_event.position.x

        if row == 1:
            if mouse_event.event_type == MouseEventType.MOUSE_UP:
                if mouse_event.button == MouseButton.LEFT and callable(
                    handler := self.mouse_handlers.get(col)
                ):
                    # Activate the tab
                    handler()
                    return None
                elif mouse_event.button == MouseButton.MIDDLE and self.closeable:
                    if callable(handler := self.mouse_handlers.get(col)):
                        # Activate tab
                        handler()
                        # Close the now active tab
                        tabs = self.tabs
                        on_close = tabs[self.active]
                        if callable(on_close):
                            on_close()
                    return None

        tabs = self.tabs
        if mouse_event.event_type == MouseEventType.SCROLL_UP:
            index = max(self.active - 1, 0)
            if index != self.active:
                if callable(deactivate := tabs[self.active].on_deactivate):
                    deactivate()
                if callable(activate := tabs[index].on_activate):
                    activate()
                return None
        elif mouse_event.event_type == MouseEventType.SCROLL_DOWN:
            index = min(self.active + 1, len(tabs) - 1)
            if index != self.active:
                if callable(deactivate := tabs[self.active].on_deactivate):
                    deactivate()
                if callable(activate := tabs[index].on_activate):
                    activate()
                return None
        return NotImplemented


class StackedSplit(metaclass=ABCMeta):
    """Base class for containers with selectable children."""

    def __init__(
        self,
        children: Sequence[AnyContainer],
        titles: Sequence[AnyFormattedText],
        active: int = 0,
        style: str | Callable[[], str] = "class:tab-split",
        on_change: Callable[[StackedSplit], None] | None = None,
        width: AnyDimension = None,
        height: AnyDimension = None,
    ) -> None:
        """Create a new tabbed container instance.

        Args:
            children: A list of child container or a callable which returns such
            titles: A list of tab titles or a callable which returns such
            active: The index of the active tab
            style: A style to apply to the tabbed container
            on_change: Callback to run when the selected tab changes
            width: The width of the split container
            height: The height of the split container
        """
        self._children = list(children)
        self._titles = list(titles)
        self._active: int | None = active
        self.style = style
        self.on_change = Event(self, on_change)
        self.width = width
        self.height = height

        self.container = self.load_container()

    @abstractmethod
    def load_container(self) -> AnyContainer:
        """Abstract method for loading the widget's container."""
        ...

    def add_style(self, style: str) -> str:
        """Add a style to the widget's base style."""
        base_style = self.style() if callable(self.style) else self.style
        return f"{base_style} {style}"

    @property
    def active(self) -> int | None:
        """Return the index of the active child container."""
        return self._active

    @active.setter
    def active(self, value: int | None) -> None:
        """Set the active child container.

        Args:
            value: The index of the tab to make active
        """
        if value is not None:
            value = max(0, min(value, len(self.children)))
        if value != self._active:
            self._active = value
            self.refresh()
            self.on_change.fire()
            if value is not None:
                try:
                    get_app().layout.focus(self.children[value])
                except ValueError:
                    pass

    @property
    def children(self) -> list[AnyContainer]:
        """Return a list of the widget's child containers."""
        return self._children

    @children.setter
    def children(self, value: Sequence[AnyContainer]) -> None:
        """Set the widget's child containers."""
        self._children = list(value)
        self.refresh()

    def active_child(self) -> AnyContainer:
        """Return the currently active child container."""
        return self.children[self.active or 0]

    @property
    def titles(self) -> list[AnyFormattedText]:
        """Return the titles of the child containers."""
        return self._titles

    @titles.setter
    def titles(self, value: Sequence[AnyFormattedText]) -> None:
        """Set the titles of the child containers."""
        self._titles = list(value)
        self.refresh()

    def refresh(self) -> None:
        """Reload the widget's container when its children or their titles change."""
        pass

    def __pt_container__(self) -> AnyContainer:
        """Return the widget's container."""
        return self.container


class TabbedSplit(StackedSplit):
    """A container which switches between children using tabs."""

    def __init__(
        self,
        children: Sequence[AnyContainer],
        titles: Sequence[AnyFormattedText],
        active: int = 0,
        style: str | Callable[[], str] = "class:tab-split",
        on_change: Callable[[StackedSplit], None] | None = None,
        width: AnyDimension = None,
        height: AnyDimension = None,
        border: GridStyle = OutsetGrid,
        show_borders: DiBool | None = None,
    ) -> None:
        """Initialize a new tabbed container."""
        self.border = border
        self.show_borders = show_borders or DiBool(False, True, True, True)
        super().__init__(
            children=children,
            titles=titles,
            active=active,
            style=style,
            on_change=on_change,
            width=width,
            height=height,
        )

    def load_container(self) -> AnyContainer:
        """Create the tabbed widget's container.

        Consists of a tab-bar control above a dynamic container which shows the active
        child container.

        Returns:
            The widget's container
        """
        self.control = TabBarControl(self.load_tabs(), active=self.active or 0)
        return HSplit(
            [
                Window(
                    self.control,
                    style=partial(self.add_style, "class:tab-bar"),
                    height=2,
                ),
                Border(
                    Box(
                        DynamicContainer(self.active_child),
                        padding=0,
                        padding_top=1,
                        padding_bottom=1,
                        style="class:tabbed-split,page",
                    ),
                    border=self.border,
                    show_borders=self.show_borders,
                ),
            ],
            style="class:tabbed-split",
            width=self.width,
            height=self.height,
        )

    def refresh(self) -> None:
        """Refresh the widget - set the tab-bar's tabs and active tab index."""
        self.control.tabs = self.load_tabs()
        if self.active is not None:
            self.control.active = self.active

    def load_tabs(self) -> list[TabBarTab]:
        """Return a list of tabs for the current children."""
        return [
            TabBarTab(
                title=title,
                on_activate=partial(setattr, self, "active", i),
            )
            for i, title in enumerate(self.titles)
        ]


class AccordionSplit(StackedSplit):
    """A container which switches between children using expandable sections."""

    def load_container(self) -> AnyContainer:
        """Create the accordiion widget's container."""
        self.draw_container()
        return DynamicContainer(lambda: self._container)

    def draw_container(self) -> None:
        """Render the accordion in it's current state."""
        self._container = HSplit(
            [
                Border(
                    HSplit(
                        [
                            Window(
                                FormattedTextControl(
                                    partial(self.title_text, index, title),
                                    focusable=True,
                                    show_cursor=False,
                                )
                            ),
                            ConditionalContainer(
                                Box(child, padding_left=0),
                                filter=Condition(
                                    partial(lambda i: self.active == i, index)
                                ),
                            ),
                        ]
                    ),
                    style=partial(self.add_style, "class:border"),
                )
                for index, (title, child) in enumerate(zip(self.titles, self.children))
            ],
            style="class:accordion",
        )

    def title_text(self, index: int, title: AnyFormattedText) -> StyleAndTextTuples:
        """Generate the title for each child container."""
        return [
            ("", " "),
            (
                "bold" + (" class:selection" if self.active == index else ""),
                "▶" if self.active == index else "▼",
                cast(
                    "Callable[[MouseEvent], None]", partial(self.mouse_handler, index)
                ),
            ),
            ("", " "),
            *[
                (
                    f"bold {style}",
                    text,
                    cast(
                        "Callable[[MouseEvent], None]",
                        partial(self.mouse_handler, index),
                    ),
                )
                for style, text, *_ in to_formatted_text(title)
            ],
        ]

    def mouse_handler(
        self, index: int, mouse_event: MouseEvent
    ) -> NotImplementedOrNone:
        """Handle mouse events."""
        # if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
        #    get_app().layout.focus()
        if mouse_event.event_type == MouseEventType.MOUSE_UP:
            self.toggle(index)
        else:
            return NotImplemented
        return None

    def toggle(
        self,
        index: int,
    ) -> None:
        """Toggle the visibility of a child container."""
        self.active = index if self.active != index else None

    def refresh(self) -> None:
        """Re-draw the container when the list of child containers changes."""
        self.draw_container()



"""Define default key-bindings."""

from __future__ import annotations

from typing import TYPE_CHECKING

from prompt_toolkit.key_binding import KeyBindings

from euporie.core.commands import get_cmd
from euporie.core.config import Config
from euporie.core.current import get_app
from euporie.core.keys import Keys

if TYPE_CHECKING:
    from prompt_toolkit.key_binding import KeyBindingsBase

    AnyKey = tuple[Keys | str, ...] | Keys | str
    AnyKeys = list[AnyKey] | AnyKey

    KeyBindingDefs = dict[str, AnyKeys]

BINDINGS: dict[str, KeyBindingDefs] = {}


def register_bindings(bindings: dict[str, KeyBindingDefs]) -> None:
    """Update the key-binding registry."""
    for group, items in bindings.items():
        if group not in BINDINGS:
            BINDINGS[group] = {}
        for command, keys in items.items():
            BINDINGS[group][command] = keys


def load_registered_bindings(
    *names: str, config: Config | None = None
) -> KeyBindingsBase:
    """Assign key-bindings to commands based on a dictionary."""
    from euporie.core.app import BaseApp

    app = get_app()
    if config is None and isinstance(app, BaseApp):
        config = app.config

    kb = KeyBindings()
    for name in names:
        binding_dict = BINDINGS.get(name, {})
        # Augment with bindings from config
        if config is not None:
            binding_dict.update(config.key_bindings.get(name, {}))
        for command, keys in binding_dict.items():
            get_cmd(command).bind(kb, keys)
    return kb



"""Define a tree-view widget."""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    HSplit,
    VSplit,
    Window,
)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.mouse_events import MouseButton, MouseEvent, MouseEventType

if TYPE_CHECKING:
    from typing import Any

    from prompt_toolkit.formatted_text import StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.layout.containers import AnyContainer


class JsonView:
    """A JSON-view container."""

    def __init__(
        self, data: Any, title: str | None = None, expanded: bool = True
    ) -> None:
        """Create a new instance."""
        if title is None:
            self.title = "root"
        else:
            self.title = title
        self.expanded = expanded

        self.children = []
        self.value_style = ""
        self.value = ""

        if isinstance(data, list):
            data = dict(enumerate(data))
            self.value = f"[] {len(data)} items"
            self.value_style = "class:pygments.comment"
        if isinstance(data, dict):
            self.value = f"{{}} {len(data)} items"
            self.value_style = "class:pygments.comment"
            self.children = [
                JsonView(data=value, title=key, expanded=expanded)
                for key, value in data.items()
            ]
        else:
            self.value = f"{data!r}"
            self.value_style = {
                str: "class:pygments.literal.string",
                int: "class:pygments.literal.number",
                float: "class:pygments.literal.number",
                bool: "class:pygments.keyword.constant",
            }.get(type(data), "")

        self.container = HSplit(
            [
                Window(
                    FormattedTextControl(self.format_title),
                    dont_extend_height=True,
                ),
                ConditionalContainer(
                    VSplit(
                        [
                            Window(
                                width=2,
                                char="  ",
                                style="",
                            ),
                            HSplit(self.children),
                        ]
                    ),
                    filter=Condition(lambda: self.expanded and bool(self.children)),
                ),
            ],
            style="class:tree",
        )

    def toggle(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Toggle the expansion state."""
        if (
            mouse_event.button == MouseButton.LEFT
            and mouse_event.event_type == MouseEventType.MOUSE_UP
        ):
            self.expanded = not self.expanded
            return None
        return NotImplemented

    def format_title(self) -> StyleAndTextTuples:
        """Return the tree node toggle and title."""
        return cast(
            "StyleAndTextTuples",
            [
                ("class:pygments.operator", "▼" if self.expanded else "▶", self.toggle)
                if self.children
                else ("", " ", self.toggle),
                ("", " ", self.toggle),
                ("class:pygments.keyword", f"{self.title}", self.toggle),
                ("class:pygments.punctuation", ": ", self.toggle),
                (self.value_style, self.value, self.toggle),
            ],
        )

    def __pt_container__(self) -> AnyContainer:
        """Return the tree-view container's content."""
        return self.container



"""Contain a tab for displaying files."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.layout.containers import VSplit
from prompt_toolkit.layout.dimension import Dimension

from euporie.core.filters import insert_mode, replace_mode
from euporie.core.kernel import Kernel, MsgCallbacks
from euporie.core.key_binding.registry import load_registered_bindings
from euporie.core.lexers import detect_lexer
from euporie.core.margins import MarginContainer, ScrollbarMargin
from euporie.core.tabs.base import KernelTab
from euporie.core.widgets.inputs import KernelInput

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Callable

    from prompt_toolkit.layout.containers import AnyContainer

    from euporie.core.app import BaseApp
    from euporie.core.comm.base import Comm
    from euporie.core.widgets.status_bar import StatusBarFields

log = logging.getLogger(__name__)


class EditorTab(KernelTab):
    """Tab class for editing text files."""

    name = "Text Editor"
    weight = 1
    mime_types = {"text/*"}

    allow_stdin = True
    _metadata = {}

    def __init__(
        self,
        app: BaseApp,
        path: Path | None = None,
        kernel: Kernel | None = None,
        comms: dict[str, Comm] | None = None,
        use_kernel_history: bool = False,
    ) -> None:
        """Call when the tab is created."""
        self.default_callbacks = MsgCallbacks({})
        self._metadata = {}
        self.loaded = False

        super().__init__(app, path, kernel, comms, use_kernel_history)

    def post_init_kernel(self) -> None:
        """Load UI and file in background after kernel has inited."""
        # Load the UI
        self.container = self.load_container()
        self.app.layout.focus(self.container)

        # Read file
        self.load()

    def load(self) -> None:
        """Load the text file."""
        if self.path is not None:
            text = self.path.read_text()
        else:
            text = ""

        # Set text
        self.input_box.text = text
        self.input_box.buffer.on_text_changed += lambda b: setattr(self, "dirty", True)
        self.input_box.read_only = False
        self.loaded = True

        # Detect language
        lexer = detect_lexer(text[:1000], self.path)
        if lexer:
            self._metadata = {"kernelspec": {"language": lexer.name}}
        # Re-lex the file
        self.input_box.control._fragment_cache.clear()
        self.app.invalidate()

    @property
    def position(self) -> str:
        """Return the position of the cursor in the document."""
        doc = self.input_box.buffer.document
        return f"{doc.cursor_position_row + 1}:{doc.cursor_position_col}"

    def __pt_status__(self) -> StatusBarFields | None:
        """Return a list of statusbar field values shown then this tab is active."""
        if not self.loaded:
            return (["Loading…"], [])
        return (
            [
                ("I" if insert_mode() else ("o" if replace_mode() else ">")),
                self.position,
            ],
            [str(self.path)],
        )

    @property
    def path_name(self) -> str:
        """Return the path name."""
        if self.path is not None:
            return self.path.name
        else:
            return "(New file)"

    @property
    def title(self) -> str:
        """Return the tab title."""
        return ("* " if self.dirty else "") + self.path_name

    def load_container(self) -> AnyContainer:
        """Abcract method for loading the notebook's main container."""
        assert self.path is not None

        self.input_box = KernelInput(kernel_tab=self, right_margins=[], read_only=True)

        return VSplit(
            [
                self.input_box,
                MarginContainer(ScrollbarMargin(), target=self.input_box.window),
            ],
            width=Dimension(weight=1),
            height=Dimension(weight=1),
            key_bindings=load_registered_bindings("euporie.core.tabs.base.Tab"),
        )

    def save(self, path: Path | None = None, cb: Callable | None = None) -> None:
        """Save the current file."""
        if path is not None:
            self.path = path

        if self.path is None:
            if dialog := self.app.dialogs.get("save-as"):
                dialog.show(tab=self, cb=cb)
        else:
            log.debug("Saving file...")
            self.saving = True
            self.app.invalidate()

            # Save to a temp file, then replace the original
            temp_path = self.path.parent / f".{self.path.stem}.tmp{self.path.suffix}"
            log.debug("Using temporary file %s", temp_path.name)
            try:
                open_file = temp_path.open("w")
            except NotImplementedError:
                if dialog := self.app.dialogs.get("save-as"):
                    dialog.show(tab=self, cb=cb)
            else:
                try:
                    open_file.write(self.input_box.buffer.text)
                except Exception:
                    if dialog := self.app.dialogs.get("save-as"):
                        dialog.show(tab=self, cb=cb)
                else:
                    try:
                        temp_path.rename(self.path)
                    except Exception:
                        if dialog := self.app.dialogs.get("save-as"):
                            dialog.show(tab=self, cb=cb)
                    else:
                        self.dirty = False
                        self.saving = False
                        self.app.invalidate()
                        log.debug("File saved")
            # Run the callback
            if callable(cb):
                cb()
